"""Module containing fundamental physics formulas.

:Author(s): **Christoffer Hjertø Grindheim**
"""
import typing as t
from numbers import Number
from collections.abc import Sequence

import numpy as np
from multipledispatch import dispatch
from scipy import optimize, constants

if t.TYPE_CHECKING:
    from ..tracking.machine import Machine
    from ..tracking.programs_machine import ProgramsMachine
    from ..tracking.machine_base import MachineABC


arr_or_float = t.Union[np.ndarray, float]


def b_to_e(machine: 'Machine') -> float:
    """
    Calculates the energy for a particle
    in a circular machine at dipole field B.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.

    Returns
    -------
    energy: float
        Energy calculated from B-field.
    """
    return np.sqrt((machine.q
                    * machine.b0
                    * machine.bending_rad
                    * constants.c) ** 2
                   + machine.e_rest ** 2)


def lorentz_beta(machine: 'MachineABC', rf_turn: int) -> float:
    """Calculates Lorentz beta factor (v/c) at a turn.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.
    rf_turn: int
        At which RF turn the calculation should happen.

    Returns
    -------
    Lorentz beta: float
        The Lorenz beta factor (v/c) at a given turn.
    """
    return np.sqrt(1.0 - (float(machine.e_rest) /
                          machine.e0[rf_turn]) ** 2)


def rfvolt_rf1(phi: float, v1: float, bdot: float, radius: float,
               bending_rad: float, charge: float) -> float:
    """Calculates RF voltage1 seen by particle.
    Needed by the Newton root finder to calculate phi0.

    Parameters
    ----------
    phi: float
        Phase where voltage is to be calculated.
    v1: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    radius: float
        Mean orbit radius of machine [m].
    bending_rad: float
        Machine bending radius [m].
    charge: float
        Charge state of accelerated particle.

    Returns
    -------
    voltage: float
        Voltage from RF system 1, as seen by particle.
    """
    q_sign = np.sign(charge)
    return (v1 * np.sin(phi)
            - 2 * np.pi * radius
            * bending_rad * bdot * q_sign)


def drfvolt_rf1(phi: arr_or_float, v1: arr_or_float) -> arr_or_float:
    """Calculates derivative of RF voltage1 seen by particle.
    Needed by the Newton root finder to calculate phi0.

    Parameters
    ----------
    phi: float
        Phase where voltage is to be calculated.
    v1: float or np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.

    Returns
    -------
    Derivative of voltage: float
        Time-derivative of voltage from RF system 1, as seen by particle.
    """
    return v1 * np.cos(phi)


def rfvolt_rf1_mch(phi: float, machine: 'Machine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    turn_time = machine.time_at_turn[rf_turn]
    v1 = vrft(machine.vrf1, machine.vrf1dot, turn_time)

    return rfvolt_rf1(phi, v1, machine.bdot, machine.mean_orbit_rad,
                      machine.bending_rad, machine.q)


def drfvolt_rf1_mch(phi: float, machine: 'Machine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    turn_time = machine.time_at_turn[rf_turn]
    v1 = vrft(machine.vrf1, machine.vrf1dot, turn_time)
    return drfvolt_rf1(phi, v1)


def rfvolt_rf1_pmch(phi: float, machine: 'ProgramsMachine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    v1 = machine.vrf1_at_turn[rf_turn]
    bdot = machine.bdot[rf_turn]

    return rfvolt_rf1(phi, v1, bdot, machine.mean_orbit_rad,
                      machine.bending_rad, machine.q)


def drfvolt_rf1_pmch(phi: float, machine: 'ProgramsMachine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    v1 = machine.vrf1_at_turn[rf_turn]
    return drfvolt_rf1(phi, v1)


def rf_voltage(phi: arr_or_float, v1: arr_or_float, v2: arr_or_float,
               bdot: arr_or_float, h_ratio: float, phi12: arr_or_float,
               radius: float, bending_rad: float, charge: float) \
        -> arr_or_float:
    """Calculates RF voltage from both RF systems seen by particle.
    Needed by the Newton root finder to calculate phi0.

    Parameters
    ----------
    phi: float
        Phase where voltage is to be calculated.
    v1: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    v2: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    h_ratio: float
        Ratio of harmonics between the two RF systems.
    phi12: float or np.ndarray
        Phase difference between the two RF systems, turn-by-turn, or
        considered constant.
    radius: float
        Mean orbit radius of machine [m].
    bending_rad: float
        Machine bending radius [m].
    charge: float
        Charge state of accelerated particle.

    Returns
    -------
    voltage: float
        Voltage from both RF systems, as seen by particle.
    """
    q_sign = np.sign(charge)
    return (v1 * np.sin(phi)
            + (v2 * np.sin(h_ratio * (phi - phi12)))
            - (np.pi * 2 * radius
               * bending_rad * bdot * q_sign))


def rf_voltage_mch(phi: float, machine: 'Machine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    turn_time = machine.time_at_turn[rf_turn]
    v1 = vrft(machine.vrf1, machine.vrf1dot, turn_time)
    v2 = vrft(machine.vrf2, machine.vrf2dot, turn_time)

    return rf_voltage(phi, v1, v2, machine.bdot, machine.h_ratio,
                      machine.phi12, machine.mean_orbit_rad,
                      machine.bending_rad, machine.q)


def rf_voltage_pmch(phi: float, machine: 'ProgramsMachine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    v1 = machine.vrf1_at_turn[rf_turn]
    v2 = machine.vrf2_at_turn[rf_turn]
    bdot = machine.bdot[rf_turn]
    phi12 = machine.phi12[rf_turn]

    return rf_voltage(phi, v1, v2, bdot, machine.h_ratio,
                      phi12, machine.mean_orbit_rad,
                      machine.bending_rad, machine.q)


def drf_voltage(phi: float, v1: arr_or_float, v2: arr_or_float,
                h_ratio: float, phi12: arr_or_float) -> arr_or_float:
    """Calculates derivative of RF voltage from both RF systems seen by
    particle.
    Needed by the Newton root finder to calculate phi0.

    Parameters
    ----------
    phi: float
        Phase where voltage is to be calculated.
    v1: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    v2: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    h_ratio: float
        Ratio of harmonics between the two RF systems.
    phi12: float or np.ndarray
        Phase difference between the two RF systems, turn-by-turn, or
        considered constant.

    Returns
    -------
    Derivative of voltage: float
        Derivative of voltage from both RF systems, as seen by particle.
    """
    return v1 * np.cos(phi) + h_ratio * v2 * np.cos(h_ratio * (phi - phi12))


def drf_voltage_mch(phi: float, machine: 'Machine', rf_turn: int) -> float:
    """Objective function used in phi0 Newton optimization"""
    turn_time = machine.time_at_turn[rf_turn]
    v1 = vrft(machine.vrf1, machine.vrf1dot, turn_time)
    v2 = vrft(machine.vrf2, machine.vrf2dot, turn_time)
    return (v1 * np.cos(phi)
            + machine.h_ratio
            * v2
            * np.cos(machine.h_ratio
                     * (phi - machine.phi12)))


def drf_voltage_pmch(phi: float, machine: 'ProgramsMachine', rf_turn: int) \
        -> float:
    """Objective function used in phi0 Newton optimization"""
    v1 = machine.vrf1_at_turn[rf_turn]
    v2 = machine.vrf2_at_turn[rf_turn]
    phi12 = machine.phi12[rf_turn]
    return (v1 * np.cos(phi) + machine.h_ratio * v2 * np.cos(machine.h_ratio
            * (phi - phi12)))


@dispatch(Number, Number, Number, Number, Number, Number, Number,
          (Sequence, np.ndarray), Number)
def rf_voltage_at_phase(phi: float, vrf1: float, vrf1dot: float, vrf2: float,
                        vrf2dot: float, h_ratio: float, phi12: float,
                        time_at_turns: np.ndarray, rf_turn: int):
    """RF voltage formula without calculating the difference in E0."""
    turn_time = time_at_turns[rf_turn]
    v1 = vrft(vrf1, vrf1dot, turn_time)
    v2 = vrft(vrf2, vrf2dot, turn_time)
    voltage = rf_voltage_at_phase(phi, v1, v2, h_ratio, phi12)
    return voltage


@dispatch(Number, Number, Number, Number, Number)
def rf_voltage_at_phase(phi: float, vrf1: float, vrf2: float, h_ratio: float,
                        phi12: float):
    voltage = vrf1 * np.sin(phi) + vrf2 * np.sin(h_ratio * (phi - phi12))
    return voltage


def vrft(vrf: float, vrf_dot: float, turn_time: float) -> float:
    """Calculates the RF peak voltage at turn rf_turn
    assuming a linear voltage function.
    Time=0 at machine reference frame.

    Parameters
    ----------
    vrf: float
        Peak voltage.
    vrf_dot: float
        Time derivative of RF voltage.
    turn_time: float
        time at turn for which the RF voltage should be calculated.

    Returns
    -------
    RF voltage: float
        RF voltage at turn.
    """
    return vrf + vrf_dot * turn_time


def find_synch_phase_mch(machine: 'Machine', rf_turn: int,
                         phi_lower: float, phi_upper: float) -> float:
    """Uses the Newton-Raphson root finder to estimate
    the synchronous phase for a particle on the normal orbit.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.
    rf_turn: int
        At which RF turn the calculation should happen.
    phi_lower:
        Lower phase boundary
    phi_upper:
        Upper phase boundary

    Returns
    -------
    synchronous phase: float
        Synchronous phase given in gradients.
    """
    try:
        phi_start = optimize.newton(func=rfvolt_rf1_mch,
                                    x0=(phi_lower + phi_upper) / 2.0,
                                    fprime=drfvolt_rf1_mch,
                                    tol=0.0001,
                                    maxiter=100,
                                    args=(machine, rf_turn))
        synch_phase = optimize.newton(func=rf_voltage_mch,
                                      x0=phi_start,
                                      fprime=drf_voltage_mch,
                                      tol=0.0001,
                                      maxiter=100,
                                      args=(machine, rf_turn))
    except RuntimeError:
        raise ValueError('Could not fit synchronous phase for the supplied '
                         'parameters')

    return synch_phase


def find_phi_lower_upper(machine: 'MachineABC', rf_turn: int) \
        -> t.Tuple[float, float]:
    """Calculates lower and upper phase of RF voltage
    for use in estimation of phi0.

    Parameters
    ----------
    machine: MachineABC
        Stores machine parameters and reconstruction settings.
    rf_turn: int
        At which RF turn the calculation should happen.

    Returns
    -------
    phi lower: float
        Lower phase boundary [rad].
    phi upper: float
        Upper phase boundary [rad].

    """
    condition = (machine.q
                 * (machine.trans_gamma
                    - (machine.e0[rf_turn]
                       / float(machine.e_rest)))) > 0
    if condition:
        phi_lower = -np.pi
        phi_upper = np.pi
    else:
        phi_lower = 0.0
        phi_upper = 2 * np.pi
    return phi_lower, phi_upper


def phase_slip_factor(machine: 'MachineABC') -> np.ndarray:
    """Calculates phase slip factor at each turn.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.

    Returns
    -------
    phase slip factor: ndarray
        1D array of phase slip factors for each turn.
    """
    return (1.0 - machine.beta0 ** 2) - machine.trans_gamma ** (-2)


def find_dphase(machine: 'MachineABC') -> np.ndarray:
    """Calculates coefficient needed for drift calculation during tracking.
    The drift coefficient is calculated for each machine turn.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.

    Returns
    -------
    drift factor: ndarray
        1D array holding drift coefficient for each machine turn.
    """
    return (2 * np.pi * machine.h_num * machine.eta0
            / (machine.e0 * machine.beta0 ** 2))


def revolution_freq(machine: 'MachineABC') -> np.ndarray:
    """Calculate revolution frequency for each turn.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.

    Returns
    -------
    revolution frequency: ndarray
        1D array holding revolution frequency at each machine turn.
    """
    return machine.beta0 * constants.c / machine.mean_orbit_rad


def calc_self_field_coeffs(machine: 'MachineABC') -> np.ndarray:
    """Calculates self-field coefficient for each profile.
    Needed for calculation of self-fields.

    Parameters
    ----------
    machine: Machine
        Stores machine parameters and reconstruction settings.

    Returns
    -------
    Self-field coefficients: ndarray
        1D array holding a self-field coefficient for each time frame.
    """
    sfc = np.zeros(machine.nprofiles)
    for i in range(machine.nprofiles):
        this_turn = i * machine.dturns
        sfc[i] = ((constants.e / machine.omega_rev0[this_turn])
                  * ((1.0 / machine.beta0[this_turn]
                      - machine.beta0[this_turn])
                     * machine.g_coupling * np.pi * 2.0e-7 * constants.c
                     - machine.zwall_over_n)
                  / machine.dtbin ** 2)
    return sfc


def phase_low(phase: float, bunch_phaselength: float, vrf1: np.ndarray,
              vrf2: np.ndarray, phi0: np.ndarray, h_ratio: float,
              phi12: arr_or_float, rf_turn: int,
              ref_frame: int = 0) -> float:
    """Calculates potential energy at phase.
    Needed for estimation of x-coordinate of synchronous particle.

    Parameters
    ----------
    phase: float
        Phase [rad] where potential energy is calculated.
    bunch_phaselength: float
        Length of bunch, given as phase [rad].
    vrf1: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    vrf2: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    phi0: ndarray
        1D array holding the synchronous phase angle at the end of each turn.
    h_ratio: float
        Ratio of harmonics between the two RF systems.
    phi12: float or np.ndarray
        Phase difference between the two RF systems, turn-by-turn, or
        considered constant.
    rf_turn: int
        Which turn to do the calculation at.
    ref_frame: int
        Frame to which machine parameters are referenced.
    """

    phi12 = phi12[ref_frame] if isinstance(phi12, np.ndarray) else phi12

    v1 = vrf1[ref_frame] * (np.cos(phase + bunch_phaselength) - np.cos(phase))

    v2 = (vrf2[ref_frame]
          * (np.cos(h_ratio
                    * (phase + bunch_phaselength - phi12))
             - np.cos(h_ratio * (phase - phi12)))
          / h_ratio)

    v_synch_phase = (bunch_phaselength
                     * rf_voltage_at_phase(phi0[rf_turn], vrf1[rf_turn],
                                           vrf2[rf_turn], h_ratio, phi12))

    return v1 + v2 + v_synch_phase


def phase_low_mch(phase: float, machine: 'MachineABC',
                  bunch_phaselength: float, rf_turn: int) -> float:
    """Calculates potential energy at phase.
    Needed for estimation of x-coordinate of synchronous particle.

    Parameters
    ----------
    phase: float
        Phase [rad] where potential energy is calculated.
    machine: Machine
        Stores machine parameters and reconstruction settings.
    bunch_phaselength: float
        Length of bunch, given as phase [rad]
    rf_turn: int
        At which RF turn the calculation should happen.
    """

    return phase_low(phase, bunch_phaselength, machine.vrf1_at_turn,
                     machine.vrf2_at_turn, machine.phi0, machine.h_ratio,
                     machine.phi12, rf_turn, machine.machine_ref_frame)


def dphase_low_mch(phase: float, machine: 'MachineABC',
                   bunch_phaselength: float, rf_turn: int, *args) -> float:
    """Calculates derivative of the phase_low function.
    The function is needed for estimation of x-coordinate
    of synchronous particle.

    Parameters
    ----------
    phase: float
        Phase [rad] where potential energy is calculated.
    machine: Machine
        Stores machine parameters and reconstruction settings.
    bunch_phaselength: float
        Length of bunch, given as phase [rad]
    rf_turn: int
        Which turn to do the calculation at.
    args: None
        Not used, arg needed for implementation of Newton-Raphson root finder.
    """
    return dphase_low(phase, bunch_phaselength, machine.vrf1_at_turn,
                      machine.vrf2_at_turn, machine.h_ratio,
                      machine.phi12, rf_turn)


def dphase_low(phase: float, bunch_phaselength: float, vrf1: np.ndarray,
                vrf2: np.ndarray, h_ratio: float, phi12: arr_or_float,
                rf_turn: int) -> float:
    """Calculates derivative of the phase_low function.
    The function is needed for estimation of x-coordinate
    of synchronous particle.

    Parameters
    ----------
    phase: float
        Phase [rad] where potential energy is calculated.
    bunch_phaselength: float
        Length of bunch, given as phase [rad].
    vrf1: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    vrf2: np.ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    h_ratio: float
        Ratio of harmonics between the two RF systems.
    phi12: float or np.ndarray
        Phase difference between the two RF systems, turn-by-turn, or
        considered constant.
    rf_turn: int
        Which turn to do the calculation at.
    """
    phi12 = phi12[rf_turn] if isinstance(phi12, np.ndarray) else phi12

    ret = (-1.0 * vrf2[rf_turn]
           * (np.sin(h_ratio
                     * (phase + bunch_phaselength - phi12))
              - np.sin(h_ratio * (phase - phi12)))
           - vrf1[rf_turn]
           * (np.sin(phase + bunch_phaselength) - np.sin(phase)))

    return ret



major_version = 3
minor_version = 4
patch_level = 0

dev_version = 8

__version__ = '{}.{}'.format(
    major_version,
    minor_version,
)

if patch_level != 0:
    __version__ += f'.{patch_level}'

if dev_version != -1:
    __version__ += '-dev{}'.format(dev_version)



class TomoException(Exception):
    pass

# ================
# Array Exceptions
# ================


class UnequalArrayShapes(TomoException):
    pass

# ====================
# Numerical Exceptions
# ====================


class OverMaxIterationsError(TomoException):
    pass


class ValuesOutOfBrackets(TomoException):
    pass


class ArrayLengthError(TomoException):
    pass


class ArrayElementsNotEqualError(TomoException):
    pass

# ===========================
# Input Parameters Exceptions
# ===========================


class InputError(TomoException):
    pass


class MachineParameterError(TomoException):
    pass


class SpaceChargeParameterError(TomoException):
    pass


class RawDataImportError(TomoException):
    pass


class ArgumentError(TomoException):
    pass


class NegativeIndexError(IndexError):
    pass

# ===========================
# Time space Exceptions
# ===========================


class RebinningError(TomoException):
    pass


class WaterfallReducedToZero(TomoException):
    pass


class FilteredProfilesError(TomoException):
    pass


class WaterfallError(TomoException):
    pass


class ProfileChargeNotCalculated(TomoException):
    pass

# ============================
# MapInfo exceptions
# ============================


class EnergyBinningError(TomoException):
    pass


class EnergyLimitsError(TomoException):
    pass


class PhaseLimitsError(TomoException):
    pass


class MapCreationError(TomoException):
    pass


# ============================
# PARTICLE TRACKING EXCEPTIONS
# ============================

class InvalidParticleError(TomoException):
    pass


class TrackingError(TomoException):
    pass


class SelfFieldTrackingError(TomoException):
    pass

# ============================
# TOMOGRAPHY EXCEPTIONS
# ============================


class PhaseSpaceReducedToZeroes(TomoException):
    pass


class XPOutOfImageWidthError(TomoException):
    pass


class CoordinateImportError(TomoException):
    pass


class CoordinateError(TomoException):
    pass



"""Module containing function for full Fortran
style tomographic reconstruction.

:Author(s): **Christoffer Hjertø Grindheim**
"""

import logging
import typing as t
import numpy as np

from ..data import data_treatment as dtreat
from ..tomography import tomography as tomography
from ..tracking import particles as pts
# Tomo modules
from ..tracking import tracking as tracking
from ..utils import tomo_input as tomoin, tomo_output as tomoout

from ..compat import tomoscope as tscp

log = logging.getLogger(__name__)


def run(input: str, reconstruct_profile: bool = None,
        output_dir: str = None, tomoscope: bool = False,
        plot: bool = False) \
        -> t.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Function to perform full reconstruction based on the original
    algorithm.

    Parameters
    ----------
    input: string
        Path to input file.
    reconstruct_profile: int, optional, default=None
        Profile to be reconstructed. If not provided, machine.filmstart
        will be reconstructed.
    output_dir: str
        Output directory when saving tomoscope output.
    tomoscope: bool
        Enable tomoscope specific output for progress tracking.
    plot: bool = False
        Plot phase space after reconstruction.

    Returns
    -------
    Phase axis: ndarray
        1D array containing time axis of reconstructed phase space image.
    Energy axis: ndarray
        1D array containing energy axis of reconstructed phase space image.
    density: ndarray
        2D array containing the reconstructed phase space image.

    Example
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> import tomo.utils.run_tomo as tomorun
    >>>
    >>> filepath = '...my/favourite/input.dat'
    >>> tRange, ERange, density = tomorun.run_file(filepath)
    >>>
    >>> vmin = np.min(density[density>0])
    >>> vmax = np.max(density)
    >>> plt.contourf(tRange*1E9, ERange/1E6, density.T,
                     levels=np.linspace(vmin, vmax, 50), cmap='Oranges')
    >>> plt.xlabel('dt (ns)')
    >>> plt.ylabel('dE (MeV)')
    >>> plt.show()
    """

    raw_params, raw_data = tomoin.get_user_input(input)

    if tomoscope:
        print(' Start')

    machine, frames = tomoin.txt_input_to_machine(raw_params)
    machine.values_at_turns()
    waterfall = frames.to_waterfall(raw_data)

    if output_dir is None or output_dir == '':
        output_dir = machine.output_dir
    else:
        log.info(f'Overriding output dir with {output_dir}')

    profiles = tomoin.raw_data_to_profiles(
        waterfall, machine, frames.rebin, frames.sampling_time)
    profiles.calc_profilecharge()

    # TODO: Insert space charge calculation from example file

    if profiles.machine.synch_part_x < 0:
        fit_info = dtreat.fit_synch_part_x(profiles)
        machine.load_fitted_synch_part_x_ftn(fit_info)

    if reconstruct_profile is None:
        reconstr_idx = machine.filmstart
    else:
        reconstr_idx = reconstruct_profile

    # Tracking...
    tracker = tracking.Tracking(machine)

    if tomoscope:
        tracker.enable_fortran_output(profiles.profile_charge)

    if tracker.self_field_flag:
        profiles.calc_self_fields()
        tracker.enable_self_fields(profiles)

    xp, yp = tracker.track(reconstr_idx)

    # Converting from physical coordinates ([rad], [eV])
    # to phase space coordinates.
    if not tracker.self_field_flag:
        xp, yp = pts.physical_to_coords(
            xp, yp, machine, tracker.particles.xorigin,
            tracker.particles.dEbin)

    # Filters out lost particles, transposes particle matrix,
    # casts to np.int32.
    xp, yp = pts.ready_for_tomography(xp, yp, machine.nbins)

    # Tomography!
    tomo = tomography.TomographyCpp(profiles.waterfall, xp, yp)
    weight = tomo.run(verbose=tomoscope)

    if tomoscope:
        for film in range(machine.filmstart - 1, machine.filmstop,
                          machine.filmstep):
            tscp.save_image(xp, yp, weight, machine.nbins, film, output_dir)

        tscp.save_difference(tomo.diff, output_dir, film)

    t_range, E_range, phase_space = dtreat.phase_space(tomo, machine,
                                                       reconstr_idx)

    # Removing (if any) negative areas.
    phase_space = phase_space.clip(0.0)
    # Normalizing phase space.
    phase_space /= np.sum(phase_space)

    if plot:
        tomoout.show(phase_space, tomo.diff, profiles.waterfall[reconstr_idx])

    return t_range, E_range, phase_space



"""Module containing functions for treatment of data.

:Author(s): **Christoffer Hjertø Grindheim**
"""

from typing import Tuple, TYPE_CHECKING, Union
from warnings import warn

import numpy as np

from tomo import exceptions as expt
from tomo.utils import physics
from ..cpp_routines import libtomo
from . import pre_process

import logging

if TYPE_CHECKING:
    from tomo.data.profiles import Profiles
    from tomo.tracking.machine import Machine
    from tomo.tomography.__tomography import TomographyABC

__all__ = ['rebin', 'fit_synch_part_x', 'phase_space']

log = logging.getLogger(__name__)


def rebin(waterfall: np.ndarray, rbn: int, dtbin: float = None,
          synch_part_x: float = None) \
        -> Union[Tuple[np.ndarray, float, float], Tuple[np.ndarray, float]]:
    """Rebin waterfall from shape (P, X) to (P, Y).
    P is the number of profiles, X is the original number of bins,
    and Y is the number of bins after the re-binning.

    The algorithm is based on the rebinning function from the
    original tomography program.

    An array of length N, rebinned with a rebin factor of R will result
    in a array of length N / R. If N is not dividable on R, the resulting
    array will have the length (N / R) + 1.

    Parameters
    ----------
    waterfall: ndarray
        2D array of raw-data shaped as waterfall. Shape: (nprofiles, nbins).
    rbn: int
        Rebinning factor.
    dtbin: float
        Size of profile bins [s]. If provided, the function
        will return the new size of the bins after rebinning.
    synch_part_x: float
        x-coordinate of synchronous particle, measured in bins.
        If provided, the function will return its updated coordinate.

    Returns
    -------
    rebinned: ndarray
        Rebinned waterfall
    dtbin: float, optional, default=None
        If a dtbin has been provided in the arguments, the
        new size of the profile bins will be returned. Otherwise,
        None will be returned.
    synch_part_x: float, optional, default=None
        If a synch_part_x has been provided in the arguments, the
        new x-coordinate of the synchronous particle in bins will be returned.
        Otherwise, None will be returned.
    """
    #    warn('The rebin function has been moved to tomo.data.pre_process')
    return pre_process.rebin(waterfall, rbn, dtbin, synch_part_x)


# Original function for finding synch_part_x
# Finds synch_part_x based on a linear fit on a reference profile.
def fit_synch_part_x(profiles: 'Profiles') -> Tuple[np.ndarray, float, float]:
    """Linear fit to estimate the phase coordinate of the synchronous
    particle. The found phase is returned as a x-coordinate of the phase space
    coordinate systems in fractions of bins. The estimation is done at
    the beam reference profile, which is set in the
    :class:`tomo.tracking.machine.Machine` object.

    Parameters
    ----------
    profiles: Profiles
        Profiles object containing waterfall and information about the
        measurement.

    Returns
    -------
    fitted_synch_part_x
        X coordinate in the phase space coordinate system of the synchronous
        particle given in bin numbers.
    lower bunch limit
        Estimation of the lower bunch limit in bin numbers.
        Needed for :func:`tomo.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.
    upper bunch limit
        Estimation of the upper bunch limit in bin numbers.
        Needed for :func:`tomo.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.

    """
    #    warn('The fit_synch_part_x function has moved to '
    #         'tomo.data.pre_process')
    return pre_process.fit_synch_part_x(profiles)


def phase_space(tomo: 'TomographyABC', machine: 'Machine',
                reconstr_idx: int = 0) \
        -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """returns time, energy and phase space density arrays from a
    reconstruction, requires the homogenous distribution to have been
    generated by the particles class.

    Parameters
    ----------
    tomo: TomographyABC
        Object holding the information about a tomographic reconstruction.
    machine: Machine
        Object holding information about machine and reconstruction parameters.
    reconstr_idx: int
        Index of profile to be reconstructed.

    Returns
    -------
    t_range: ndarray
        1D array containing time axis of reconstructed phase space image.
    E_range: ndarray
        1D array containing energy axis of reconstructed phase space image.
    density: ndarray
        2D array containing the reconstructed phase space image.

    Raises
    ------
    InputError: Exception
        phase_space function requires automatic phase space generation
        to have been used.

    """
    if machine.dEbin is None:
        raise expt.InputError("""phase_space function requires automatic
                              phase space generation to have been used""")

    density = libtomo.make_phase_space(tomo.xp[:, reconstr_idx],
                                       tomo.yp[:, reconstr_idx],
                                       tomo.weight, machine.nbins)

    t_cent = machine.synch_part_x
    E_cent = machine.synch_part_y

    t_range = (np.arange(machine.nbins) - t_cent) * machine.dtbin
    E_range = (np.arange(machine.nbins) - E_cent) * machine.dEbin

    return t_range, E_range, density


# Creates a [nbins, nbins] array populated with the weights of each test
# particle
def _make_phase_space(xp: np.ndarray, yp: np.ndarray, weights: np.ndarray,
                      nbins: int) -> np.ndarray:
    log.warning('tomo.data.data_treatment._make_phase_space has moved to '
                'the C++ library at '
                'tomo.cpp_routines.libtomo.make_phase_space. '
                'This function is provided for backwards compatibility '
                'and can be removed without further notice.')
    return libtomo.make_phase_space(xp, yp, weights, nbins)


def calc_baseline_ftn(*args):
    #    warn('This function has moved to tomo.compat.fortran.')
    from tomo.compat.fortran import calc_baseline
    return calc_baseline(*args)



"""Module containing Machine class for storing
machine and reconstruction parameters

:Author(s): **Christoffer Hjertø Grindheim**
"""

import logging
from typing import Tuple

import numpy as np
from scipy import optimize, constants

from .. import assertions as asrt
from ..utils import physics
from .machine_base import MachineABC

log = logging.getLogger(__name__)

_machine_opts_def = {
    'vrf1dot': 0.0,
    'vrf2': 0.0,
    'vrf2dot': 0.0,
    'bdot': 0.0,
}


class Machine(MachineABC):
    """Class holding machine and reconstruction parameters.
    See superclass for documentation about inherited class variables.

    To summarize, the Machine class must be used if a program resembling the
    original Fortran version is to be created.

    Parameters
    ----------
    dturns: int
        Number of machine turns between each measurement.
    vrf1: float
        Peak voltage of the first RF system at the machine reference frame.
    mean_orbit_rad: float
        Mean orbit radius of machine [m].
    bending_rad: float
        Machine bending radius [m].
    b0: float
        B-field at machine reference frame [T].
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    trans_gamma: float
        Transitional gamma.
    e_rest: float
        Rest energy of accelerated particle [eV/C^2], saved as e_rest.
    nprofiles: int
        Number of measured profiles.
    nbins: int
        Number of bins in a profile.
    synch_part_x: float
        Synchronous phase given in number of bins, counting\
        from the lower profile bound to the synchronous phase.
    dtbin: float
        Size of profile bins [s].
    kwargs:
        All scalar attributes can be set via the kwargs.

    Attributes
    ----------
    vrf1: float
        Peak voltage of the first RF system at the machine reference frame.
    vrf2: float, default=0.0
        Peak voltage of the second RF system at the machine reference frame.\n
    vrf1dot: float, default=0.0
        Time derivatives of the voltages of the first RF system
        (considered constant).\n
    vrf2dot: float, default=0.0
        Time derivatives of the voltages of the second RF system
        (considered constant).\n
    b0: float
        B-field at machine reference frame [T].
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    h_num: int, default=1
        Principle harmonic number.\n
    """

    def __init__(self, dturns: int, vrf1: float, mean_orbit_rad: float,
                 bending_rad: float, b0: float, trans_gamma: float,
                 rest_energy: float, nprofiles: int, nbins: int, dtbin: float,
                 vat_now: bool = True, **kwargs):
        super().__init__(dturns, mean_orbit_rad, bending_rad, trans_gamma,
                         rest_energy, nprofiles, nbins, dtbin, **kwargs)

        kwargs_processed = super()._process_kwargs(_machine_opts_def, kwargs)

        # TODO: Take rfv info as a single input
        # TODO: Take b-field info as a single input

        # Machine parameters
        self.vrf1 = vrf1
        self.vrf1dot = kwargs_processed['vrf1dot']
        self.vrf2 = kwargs_processed['vrf2']
        self.vrf2dot = kwargs_processed['vrf2dot']
        self.b0 = b0
        self.bdot = kwargs_processed['bdot']

        self.fitted_synch_part_x = None
        self.bunchlimit_low = None
        self.bunchlimit_up = None

        if vat_now:
            self.values_at_turns()

    def values_at_turns(self):
        asrt.assert_machine_input(self)
        all_turns = (self.nprofiles - 1) * self.dturns

        # Create all-zero arrays of size nturns+1
        self._init_arrays(all_turns)

        # Calculate initial values at the machine reference frame (i0).
        i0 = self._array_initial_values()

        # Calculate remaining values for every machine turn.
        for i in range(i0 + 1, all_turns + 1):
            self.time_at_turn[i] = (self.time_at_turn[i - 1]
                                    + 2 * np.pi * self.mean_orbit_rad
                                    / (self.beta0[i - 1] * constants.c))

            try:
                self.phi0[i] = optimize.newton(func=physics.rf_voltage_mch,
                                               x0=self.phi0[i - 1],
                                               fprime=physics.drf_voltage_mch,
                                               tol=0.0001,
                                               maxiter=100,
                                               args=(self, i))
            except RuntimeError:
                raise ValueError('Could not fit synchronous phase for the '
                                 'supplied parameters.')

            self.e0[i] = (self.e0[i - 1]
                          + self.q
                          * physics.rf_voltage_at_phase(
                        self.phi0[i], self.vrf1, self.vrf1dot,
                        self.vrf2, self.vrf2dot, self.h_ratio,
                        self.phi12, self.time_at_turn, i))

            self.beta0[i] = np.sqrt(
                1.0 - (self.e_rest / float(self.e0[i])) ** 2)
            self.deltaE0[i] = self.e0[i] - self.e0[i - 1]
        for i in range(i0 - 1, -1, -1):
            self.e0[i] = (self.e0[i + 1]
                          - self.q
                          * physics.rf_voltage_at_phase(
                        self.phi0[i + 1], self.vrf1, self.vrf1dot,
                        self.vrf2, self.vrf2dot, self.h_ratio,
                        self.phi12, self.time_at_turn, i + 1))

            self.beta0[i] = np.sqrt(1.0 - (self.e_rest / self.e0[i]) ** 2)
            self.deltaE0[i] = self.e0[i + 1] - self.e0[i]

            self.time_at_turn[i] = (self.time_at_turn[i + 1]
                                    - 2 * np.pi * self.mean_orbit_rad
                                    / (self.beta0[i] * constants.c))

            try:
                self.phi0[i] = optimize.newton(func=physics.rf_voltage_mch,
                                               x0=self.phi0[i + 1],
                                               fprime=physics.drf_voltage_mch,
                                               tol=0.0001,
                                               maxiter=100,
                                               args=(self, i))
            except RuntimeError:
                raise ValueError('Could not fit synchronous phase for the '
                                 'supplied parameters.')

        # Calculate phase slip factor at each turn
        self.eta0 = physics.phase_slip_factor(self)

        # Calculates dphase for each turn
        self.drift_coef = physics.find_dphase(self)

        # Calculate revolution frequency at each turn
        self.omega_rev0 = physics.revolution_freq(self)

        # Calculate RF-voltages at each turn
        self.vrf1_at_turn, self.vrf2_at_turn = self._rfv_at_turns()

    def load_fitted_synch_part_x_ftn(self,
                                     fit_info: Tuple[float, float, float]):
        """Function for setting the synch_part_x if a fit has been performed.
        Saves parameters retrieved from the fitting routine
        needed by the :func:`tomo.compat.fortran.write_plotinfo`
        function in the :mod:`tomo.utils.tomo_output`. All needed info
        will be returned from the
        :func:`tomo.data.data_treatment.fit_synch_part_x` function.

        Sets the following fields:

        * fitted_synch_part_x
            The new x-coordinate of the synchronous particle
            (needed for :func:`tomo.cimpat.fortran.write_plotinfo`).
        * bunchlimit_low
            Lower phase of bunch (needed for
            :func:`tomo.compat.fortran.write_plotinfo`).
        * bunchlimit_up
            Upper phase of bunch (needed for
            :func:`tomo.compat.fortran.write_plotinfo`).
        * synch_part_x
            The x-coordinate of the synchronous particle.

        Parameters
        ----------
        fit_info: tuple
            Tuple should hold the following info in the following format:
            (F, L, U), where F is the fitted value of the synchronous particle,
            L is the lower bunch limit, and U is the upper bunch limit. All
            of the values should be given in bins. The info needed by the
            :func:`tomo.compat.fortran.write_plotinfo` function
            if a fit has been performed, and the a Fortran style output is
            to be given during the particle tracking.
        """
        log.info('Saving fitted synch_part_x to machine object.')
        self.fitted_synch_part_x = fit_info[0]
        self.bunchlimit_low = fit_info[1]
        self.bunchlimit_up = fit_info[2]
        self.synch_part_x = self.fitted_synch_part_x

    # Initiating arrays in order to store information about parameters
    # that has a different value every turn.
    def _init_arrays(self, all_turns: int):
        array_length = all_turns + 1
        self.time_at_turn = np.zeros(array_length)
        self.omega_rev0 = np.zeros(array_length)
        self.phi0 = np.zeros(array_length)
        self.drift_coef = np.zeros(array_length)
        self.deltaE0 = np.zeros(array_length)
        self.beta0 = np.zeros(array_length)
        self.eta0 = np.zeros(array_length)
        self.e0 = np.zeros(array_length)

    # Calculating start-values for the parameters that changes for each turn.
    # The reference frame where the start-values
    # are calculated is the machine reference frame.
    def _array_initial_values(self) -> int:
        i0 = self.machine_ref_frame * self.dturns
        self.time_at_turn[i0] = 0
        self.e0[i0] = physics.b_to_e(self)
        self.beta0[i0] = physics.lorentz_beta(self, i0)
        phi_lower, phi_upper = physics.find_phi_lower_upper(self, i0)
        # Synchronous phase of a particle on the nominal orbit
        self.phi0[i0] = physics.find_synch_phase_mch(
            self, i0, phi_lower, phi_upper)
        return i0

    # Using a linear approximation to calculate the RF voltage for each turn.
    def _rfv_at_turns(self) -> Tuple[np.ndarray, np.ndarray]:
        rf1v = self.vrf1 + self.vrf1dot * self.time_at_turn
        rf2v = self.vrf2 + self.vrf2dot * self.time_at_turn
        return rf1v, rf2v



"""Module containing the Tomography super class

:Author(s): **Christoffer Hjertø Grindheim**
"""

import logging

from abc import ABC, abstractmethod
import typing as t
import numpy as np

from tomo import exceptions as expt

log = logging.getLogger(__name__)


class TomographyABC(ABC):
    """Base class for tomography classes.

    This class holds tomography utilities and assertions.

    Parameters
    ----------
    waterfall: ndarray
        2D array of measured profiles, shaped: (nprofiles, nbins).
    x_coords: ndarray: int
        x-coordinates of particles, given as coordinates of the reconstructed
        phase space coordinate system. Shape: (nparts, nprofiles).

    Attributes
    ----------
    nparts: int
        Number of test particles.
    nprofs: int
        Number of profiles (time frames).
    nbins: int
        Number of bins in each profile.
    waterfall: ndarray
        2D array of measured profiles, shaped: (nprofiles, nbins).
        Negative values of waterfall is set to zero, and the waterfall is
        normalized.
    xp: ndarray
        x-coordinates of particles, given as coordinates of the reconstructed
        phase space coordinate system. Shape: (nparts, nprofiles).
    recreated: ndarray
        Recreated waterfall. Directly comparable with
        *Tomography.waterfall*. Shape: (nprofiles, nbins).
    diff: ndarray
        Discrepancy for phase space reconstruction at each iteration
        of the reconstruction process.
    """

    def __init__(self, waterfall: np.ndarray,
                 x_coords: np.ndarray = None, y_coords: np.ndarray = None):
        self._waterfall = self._normalize_profiles(waterfall.clip(0.0))

        self._nprofs: int = self.waterfall.shape[0]
        self._nbins: int = self.waterfall.shape[1]

        self.xp = x_coords
        self.yp = y_coords

        self.recreated = np.zeros(self.waterfall.shape)
        self.diff: np.ndarray = None
        self.weight: np.ndarray = None

    @property
    def waterfall(self) -> np.ndarray:
        """Waterfall defined as @property.

        Returns
        -------
        waterfall: ndarray
            Measured profiles (nprofiles, nbins).
            waterfall is normalized and its negative values are set to zero.
        """
        return self._waterfall

    @property
    def yp(self) -> np.ndarray:
        """Y-coordinates defined as @property.

        Parameters
        ----------
        value: ndarray, None
            2D array of tracked particles energy coordinates,
            given in phase space coordinates as integers.
            Shape: (nparts, nprofiles).

            By setting tomography.yp = None, the saved y-coordinates
            are deleted.

        Returns
        -------
        xp: ndarray
            particles y-coordinates, given as coordinates of the reconstructed
            phase space coordinate system in integers.
            shape: (nparts, nprofiles).

        Raises
        ------
        CoordinateImportError: Exception
            Provided coordinates are invalid.
        """
        return self._yp

    @yp.setter
    def yp(self, value: np.ndarray):
        if hasattr(value, '__iter__'):
            if self._xp is None:
                raise expt.CoordinateImportError(
                    'The object x-coordinates are None. x-coordinates'
                    'must be provided before the y-coordinates.')
            elif value.shape == self.xp.shape:
                self._yp = value.astype(np.int32)
            else:
                raise expt.CoordinateImportError(
                    'The given y-coordinates should be of the '
                    'same shape as the x-coordinates.')
        elif value is None:
            self._yp = None
        else:
            raise expt.CoordinateImportError(
                'Y-coordinates should be iterable, or None.')

    @property
    def xp(self) -> np.ndarray:
        """X-coordinates defined as @property.

        Automatically updates `nparts`.

        Parameters
        ----------
        value: ndarray, None
            2D array of tracked particles phase coordinates,
            given in phase space coordinates as integers.
            Shape: (nparts, nprofiles).

            By setting tomography.xp = None, the saved x-coordinates
            are deleted.

        Returns
        -------
        xp: ndarray
            particles x-coordinates, given as coordinates of the reconstructed
            phase space coordinate system in integers.
            shape: (nparts, nprofiles).

        Raises
        ------
        CoordinateImportError: Exception
            Provided coordinates are invalid.
        XPOutOfImageWidthError: Exception
            Particle(s) out of image width.
        """
        return self._xp

    @xp.setter
    def xp(self, value: np.ndarray):
        if hasattr(value, '__iter__'):
            value = np.array(value)
            if not value.ndim == 2:
                msg = 'X coordinates have two dimensions ' \
                      '(nparticles, nprofiles)'
                raise expt.CoordinateImportError(msg)
            if not value.shape[1] == self.nprofs:
                msg = f'Imported particles should be ' \
                      f'tracked trough {self.nprofs} profiles. ' \
                      f'Given particles seems so have been tracked trough ' \
                      f'{value.shape[1]} profiles.'
                raise expt.CoordinateImportError(msg)
            if np.any(value < 0) or np.any(value >= self.nbins):
                msg = 'X coordinate of particles outside of image width'
                raise expt.XPOutOfImageWidthError(msg)
            self._xp = np.ascontiguousarray(value).astype(np.int32)
            self._nparts = self._xp.shape[0]
            log.info(f'X coordinates of shape {self.xp.shape} loaded.')
        elif value is None:
            self._xp = None
            self._nparts = None
            log.info('X coordinates set to None')
        else:
            msg = 'X coordinates should be iterable, or None.'
            raise expt.CoordinateImportError(msg)

    @property
    def nparts(self) -> int:
        """Number of particles defined as @property.

        Returns
        -------
        nparts: int
            Number of test particles.
        """
        return self._nparts

    @property
    def nbins(self) -> int:
        """Number of profile bins defined as @property.

        Returns
        -------
        nbins: int
            Number of bins in each profile.
        """
        return self._nbins

    @property
    def nprofs(self) -> int:
        """Number of profiles defined as @property.

        nprofs: int
            Number of profiles (time frames).
        """
        return self._nprofs

    def _normalize_profiles(self, waterfall: np.ndarray) -> np.ndarray:
        if not waterfall.any():
            raise expt.WaterfallReducedToZero()
        waterfall /= np.sum(waterfall, axis=1)[:, None]
        return waterfall

    # Calculates discrepancy for the whole waterfall
    def _discrepancy(self, diff_waterfall: np.ndarray):
        return np.sqrt(
            np.sum(diff_waterfall ** 2) / (self.nbins * self.nprofs))

    # Created xp array modified to point at flattened version of waterfall.
    def _create_flat_points(self) -> np.ndarray:
        flat_points = self.xp.copy()
        for i in range(self.nprofs):
            flat_points[:, i] += self.nbins * i
        return flat_points

    # Finding the reciprocal of the number of particles in
    # a bin. Done to counterbalance the different amount of
    # particles in the different bins.
    # Bins with fewer particles have a larger amplification,
    # relative to bins containing many particles.
    def _reciprocal_particles(self) -> np.ndarray:
        ppb = np.zeros((self.nbins, self.nprofs))
        ppb = self._count_particles_in_bins(
            ppb, self.nprofs, self.xp, self.nparts)

        # Setting bins with zero particles one to avoid division by zero.
        ppb[ppb == 0] = 1
        return np.max(ppb) / ppb

    # Needed by reciprocal particles function.
    # TODO: removed njit, reimplement in C in the future
    def _count_particles_in_bins(self, ppb: np.ndarray,
                                 profile_count: int,
                                 xp: np.ndarray, nparts: int) -> np.ndarray:
        for i in range(profile_count):
            for j in range(nparts):
                ppb[xp[j, i], i] += 1
        return ppb

    @abstractmethod
    def run(self, niter: int = 20, verbose: bool = False,
            callback: t.Callable = None) -> np.ndarray:
        pass



"""Module containing functions for handling output from tomography programs.

Every function ending on 'ftn' creates an
output equal original Fortran program.

:Author(s): **Christoffer Hjertø Grindheim**
"""
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
from warnings import warn

from ..cpp_routines import libtomo
from ..compat import fortran


def save_profile_ftn(*args, **kwargs):
    """Here for backwards compatibility. See
    :func:`~tomo.compat.fortran.save_profile`"""
#    warn('This function has been moved to tomo.compat.fortran')
    return fortran.save_profile(*args, **kwargs)


def save_self_volt_profile_ftn(*args, **kwargs):
    """Here for backwards compatibility. See
    :func:`~tomo.compat.fortran.save_self_volt_profile`"""
#    warn('This function has been moved to tomo.compat.fortran')
    return fortran.save_self_volt_profile(*args, **kwargs)


def save_phase_space_ftn(*args, **kwargs):
    """Here for backwards compatibility. See
    :func:`~tomo.compat.fortran.save_phase_space`"""
#    warn('This function has been moved to tomo.compat.fortran')
    return fortran.save_phase_space(*args, **kwargs)


def write_plotinfo_ftn(*args, **kwargs):
    """Here for backwards compatibility. See
    :func:`~tomo.compat.fortran.write_plotinfo`"""
#    warn('This function has been moved to tomo.compat.fortran')
    return fortran.write_plotinfo(*args, **kwargs)


def save_difference_ftn(*args, **kwargs):
    """Here for backwards compatibility. See
    :func:`~tomo.compat.fortran.save_difference`"""
#    warn('This function has been moved to tomo.compat.fortran')
    return fortran.save_difference(*args, **kwargs)


def print_tracking_status_ftn(*args, **kwargs):
    """Here for backwards compatibility. See
    :func:`~tomo.compat.fortran.print_tracking_status`"""
#    warn('This function has been moved to tomo.compat.fortran')
    return fortran.print_tracking_status(*args, **kwargs)


def create_phase_space_image(
        xp: np.ndarray, yp: np.ndarray, weight: np.ndarray, n_bins: int,
        recprof: int) -> np.ndarray:
    """Convert from weighted particles to phase-space image.

    The output is equal to the phase space image created
    in the original version.

    Parameters
    ----------
    xp: ndarray
        2D array containing the x coordinates of every
        particle at every time frame. Must be given in coordinates
        of the phase space coordinate system as integers.
        Shape: (N, M), where N is the number of particles and
        M is the number of profiles.
    yp: ndarray
        2D array containing the y coordinates of every
        particle at every time frame. Must be given in coordinates
        of the phase space coordinate system as integers.
        Shape: (N, M), where N is the number of particles and
        M is the number of profiles.
    weight: ndarray
        1D array containing the weight of each particle.
    n_bins: int
        Number of bins in a profile measurement.
    recprof: int
        Index of reconstructed profile.

    Returns
    -------
    phase_space: ndarray
        Phase space presented as 2D array with shape (N, N),
        where N is the number of bins in a profile. This
        phase space image has the same format as from the original program.
    """

    phase_space = libtomo.make_phase_space(xp[:, recprof].astype(np.int32),
                                           yp[:, recprof].astype(np.int32),
                                           weight, n_bins)

    # Removing (if any) negative areas.
    phase_space = phase_space.clip(0.0)
    # Normalizing phase space.
    phase_space /= np.sum(phase_space)
    return phase_space


# --------------------------------------------------------------- #
#                         END PRODUCT                             #
# --------------------------------------------------------------- #

def show(image: np.ndarray, diff: np.ndarray, rec_prof: np.ndarray,
         figure: plt.figure = None):
    """Nice presentation of reconstruction.

    Parameters
    ----------
    image: ndarray
        Recreated phase-space image.
        Shape: (N, N), where N is the number of profile bins.
    diff: ndarray
        1D array containing discrepancies for each iteration of reconstruction.
    rec_prof: ndarray
        1D array containing the measured profile to be reconstructed.
    """

    # Normalizing recprof:
    rec_prof[:] /= np.sum(rec_prof)

    # Creating plot
    gs = gridspec.GridSpec(4, 4)

    if figure is not None:
        fig = figure
    else:
        fig = plt.figure()

    img = fig.add_subplot(gs[1:, :3])
    profs1 = fig.add_subplot(gs[0, :3])
    profs2 = fig.add_subplot(gs[1:4, 3])
    convg = fig.add_subplot(gs[0, 3])

    cimg = img.imshow(image.T, origin='lower',
                      interpolation='nearest', cmap='hot')

    profs1.plot(np.sum(image, axis=1), label='reconstructed', zorder=5)
    profs1.plot(rec_prof, label='measured', zorder=0)
    profs1.legend()

    profs2.plot(np.sum(image, axis=0),
                np.arange(image.shape[0]))

    convg.plot(diff, label='discrepancy')
    convg.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    convg.legend()

    for ax in (profs1, profs2, convg):
        ax.set_xticks([])
        ax.set_yticks([])

    convg.set_xticks(np.arange(len(diff)))
    convg.set_xticklabels([])

    if figure is None:
        fig.set_size_inches(8, 8)
        fig.tight_layout()

        plt.show()



import typing as t
import numpy as np

from .tracking import Tracking
from .tracking.machine_base import MachineABC
from .tracking import particles
from .tomography import tomography


__all__ = ['track', 'tomogram']


def track(machine: MachineABC, reconstruction_idx: int = None,
          callback: t.Callable = None) \
        -> t.Tuple[np.ndarray, np.ndarray]:

    tracker = Tracking(machine)
    dphi, denergy = tracker.track(reconstruction_idx, callback=callback)
    xp, yp = particles.physical_to_coords(dphi, denergy, machine,
                                          tracker.particles.xorigin,
                                          tracker.particles.dEbin)
    xp, yp = particles.ready_for_tomography(xp, yp, machine.nbins)

    return xp, yp


def tomogram(waterfall: np.ndarray, xp: np.ndarray, yp: np.ndarray,
             n_iter: int, callback: t.Callable = None) \
        -> tomography.Tomography:

    tomo = tomography.Tomography(waterfall, xp, yp)
    tomo.run(n_iter, callback=callback)

    return tomo



import typing as t
from abc import ABC, abstractmethod
import numpy as np
import logging

log = logging.getLogger(__name__)

_machine_base_defaults = {
    'demax': -1.E6,
    'phi12': 0.0,
    'h_ratio': 1.0,
    'synch_part_x': -1,
    'charge': 1,
    'h_num': 1,
    'g_coupling': None,
    'zwall_over_n': None,
    'min_dt': None,
    'max_dt': None,
    'self_field_flag': False,
    'full_pp_flag': False,
    'pickup_sensitivity': None,
    'machine_ref_frame': 0,
    'beam_ref_frame': 0,
    'snpt': 4,
    'niter': 20,
    'filmstart': 0,
    'filmstop': 1,
    'filmstep': 1,
    'output_dir': None,
}


class MachineABC(ABC):
    """Class holding machine and reconstruction parameters.

    ** NB: This is an abstract base class providing a skeleton for different
    implementations of the Fortran Machine class. **

    This class holds machine parameters and information about the measurements.
    Also, it holds settings for the reconstruction process.

    The Machine class and its values are needed for the original particle
    tracking routine. Its values are used for calculation of reconstruction
    area and info concerning the phase space, the distribution of particles,
    and the tracking itself. In addition to this, the machine object is needed
    for the generation of :class:`~tomo.data.profiles.Profiles` objects.

    To summarize, the Machine class must be used if a program resembling the
    original Fortran version is to be created.

    Parameters
    ----------
    dturns: int
        Number of machine turns between each measurement.
    vrf1: float
        Peak voltage of the first RF system at the machine reference frame.
    mean_orbit_rad: float
        Mean orbit radius of machine [m].
    bending_rad: float
        Machine bending radius [m].
    b0: float
        B-field at machine reference frame [T].
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    trans_gamma: float
        Transitional gamma.
    e_rest: float
        Rest energy of accelerated particle [eV/C^2], saved as e_rest.
    nprofiles: int
        Number of measured profiles.
    nbins: int
        Number of bins in a profile.
    synch_part_x: float
        Synchronous phase given in number of bins, counting\
        from the lower profile bound to the synchronous phase.
    dtbin: float
        Size of profile bins [s].
    kwargs:
        All scalar attributes can be set via the kwargs.

    Attributes
    ----------
    demax: float, default=-1.E6
        Maximum energy of reconstructed phase space.\n
    dturns: int
        Number of machine turns between each measurement.
    vrf1: float
        Peak voltage of the first RF system at the machine reference frame.
    vrf2: float, default=0.0
        Peak voltage of the second RF system at the machine reference frame.\n
    vrf1dot: float, default=0.0
        Time derivatives of the voltages of the first RF system
        (considered constant).\n
    vrf2dot: float, default=0.0
        Time derivatives of the voltages of the second RF system
        (considered constant).\n
    mean_orbit_rad: float
        Mean orbit radius of machine [m].
    bending_rad: float
        Machine bending radius [m].
    b0: float
        B-field at machine reference frame [T].
    bdot: float
        Time derivative of B-field (considered constant) [T/s].
    phi12: float, default=0.0
        Phase difference between the two RF systems (considered constant).\n
    h_ratio: float, default=1.0
        Ratio of harmonics between the two RF systems.\n
    h_num: int, default=1
        Principle harmonic number.\n
    trans_gamma: float
        Transitional gamma.
    e_rest: float
        Rest energy of accelerated particle [eV/C^2].
    q: int, default=1
        Charge state of accelerated particle.\n
    g_coupling: float, default=None
        Space charge coupling coefficient (geometrical coupling coefficient).\n
    zwall_over_n: float, default=None
        Magnitude of Zwall/n, reactive impedance.\n
    min_dt: float, default=None
        Minimum phase of reconstruction area measured in seconds.\n
    max_dt: float, default=None
        Maximum phase of reconstruction area measured in seconds.\n
    nprofiles: int
        Number of measured profiles.
    pickup_sensitivity: float, default=None
        Effective pick-up sensitivity
        (in digitizer units per instantaneous Amp).\n
    nbins: int
        Number of bins in a profile.
    synch_part_x: float
        Synchronous phase given in number of bins, counting
        from the lower profile bound to the synchronous phase.
    synch_part_y: float
        Energy coordinate of synchronous particle in phase space coordinates
        of bins. The coordinate is set to be one half of the image width
        (nbins).
    dtbin: float
        Size of profile bins [s].
    dEbin: float
        Size of profile bins in energy.
    self_field_flag: boolean, default=False
        Flag to include self-fields in the tracking.\n
    full_pp_flag: boolean, default=False
        If set, all pixels in reconstructed phase space will be tracked.\n
    machine_ref_frame: int, default=0
        Frame to which machine parameters are referenced.\n
    beam_ref_frame: int, default=0
        Frame to which beam parameters are referenced.\n
    snpt: int, default=4
        Square root of particles pr. cell of phase space.\n
    niter: int, default=20
        Number of iterations in tomographic reconstruction.\n
    filmstart: int, default=0
        First profile to reconstruct.\n
    filmstop: int, default=1
        Last profile to reconstruct.\n
    filmstep: int, default=1
        Step between profiles to reconstruct.\n
    output_dir: string, default=None
        Directory to save output.\n
    time_at_turn: ndarray
        1D array holding the time at each turn. Turn zero = 0 [s].
    phi0: ndarray
        1D array holding the synchronous phase angle at the end of each turn.
    e0: ndarray
        1D array holding the total energy of synchronous
        particle at the end of each turn.
    beta0: ndarray
        1D array holding the Lorenz beta factor (v/c)
        at the end of each turn.
    deltaE0: ndarray
        1D array holding the difference between
        e0(n) and e0(n-1) for each turn.
    eta0: ndarray
        1D array holding the phase slip factor at each turn.
    drift_coef: ndarray
        1D array holding coefficient used for calculating difference,
        from phase n to phase n + 1. Needed by trajectory height
        calculator, and tracking.
    omega_rev0: ndarray
        1D array holding the revolution frequency at each turn.
    vrf1_at_turn: ndarray
        1D array holding the peak voltage at each turn for
        the first RF station.
    vrf2_at_turn: ndarray
        1D array holding the peak voltage at each turn for
        the second RF station.
    """

    def __init__(self, dturns: int, mean_orbit_rad: float, bending_rad: float,
                 trans_gamma: float, rest_energy: float,
                 n_profiles: int, n_bins: int, dtbin: float, **kwargs):
        processed_kwargs = self._process_kwargs(_machine_base_defaults, kwargs)

        self.dturns: int = dturns
        self.mean_orbit_rad: float = mean_orbit_rad
        self.bending_rad: float = bending_rad
        self.trans_gamma: float = trans_gamma
        self.e_rest: float = rest_energy

        self.nprofiles: int = n_profiles
        self.nbins: int = n_bins

        self.dtbin = dtbin

        # kwargs
        if processed_kwargs['min_dt'] is not None:
            self.min_dt = processed_kwargs['min_dt']
        else:
            self.min_dt = 0.0

        if processed_kwargs['max_dt'] is not None:
            self.max_dt = processed_kwargs['max_dt']
        else:
            self.max_dt = n_bins * dtbin

        self.demax: float = processed_kwargs['demax']
        self.phi12: t.Union[float, np.ndarray] = processed_kwargs['phi12']
        self.h_ratio: float = processed_kwargs['h_ratio']
        self.h_num: int = processed_kwargs['h_num']
        self.q: float = processed_kwargs['charge']
        self.g_coupling: float = processed_kwargs['g_coupling']
        self.zwall_over_n: float = processed_kwargs['zwall_over_n']
        self.synch_part_x: float = processed_kwargs['synch_part_x']

        # kwargs f:lags
        self.self_field_flag: bool = processed_kwargs['self_field_flag']
        self.full_pp_flag: bool = processed_kwargs['full_pp_flag']

        # kwargs reconstruction parameters
        self.machine_ref_frame: int = processed_kwargs['machine_ref_frame']
        self.beam_ref_frame: int = processed_kwargs['beam_ref_frame']
        self.pickup_sensitivity = processed_kwargs['pickup_sensitivity']
        self.snpt: int = processed_kwargs['snpt']
        self.niter: int = processed_kwargs['niter']
        self.filmstart: int = processed_kwargs['filmstart']
        self.filmstop: int = processed_kwargs['filmstop']
        self.filmstep: int = processed_kwargs['filmstep']
        self.output_dir: str = processed_kwargs['output_dir']

        # initialise attributes for later use
        # values at turns
        self.phi0: np.ndarray = None
        self.eta0: np.ndarray = None
        self.drift_coef: np.ndarray = None
        self.deltaE0: np.ndarray = None
        self.beta0: np.ndarray = None
        self.e0: np.ndarray = None
        self.omega_rev0: np.ndarray = None
        self.time_at_turn: np.ndarray = None
        self.vrf1_at_turn: np.ndarray = None
        self.vrf2_at_turn: np.ndarray = None

        # Used as flag for checking if particles particle tracking
        # has been done
        self.dEbin: float = None

    @property
    def nbins(self) -> int:
        """nbins defined as @property.
        Updates the position of the y-coordinate of the synchronous
        particle in the phase space coordinate system when set.

        Parameters
        ----------
        nbins: int
            Number of bins in a profile.

        Returns
        -------
        nbins: int
            Number of bins in a profile.
        """
        return self._nbins

    @nbins.setter
    def nbins(self, in_nbins: int):
        self._nbins = in_nbins
        self.synch_part_y = self._nbins / 2.0
        log.debug(f'synch_part_y was updated when the '
                  f'number of profile bins changed.\nNew values - '
                  f'nbins: {self._nbins}, synch_part_y: {self.synch_part_y}')

    # Function for asserting input dictionary for machine creator
    @classmethod
    def _process_kwargs(cls, defaults: t.Dict[str, t.Any], kwargs) -> t.Dict:
        use_params = {}

        for key in defaults:
            use_params[key] = defaults[key]

        for key in kwargs.copy():
            if key not in defaults:
                pass
                # raise KeyError(f'{key} is not a machine parameter')
            else:
                use_params[key] = kwargs.pop(key)
        return use_params

    @abstractmethod
    def values_at_turns(self):
        """Calculating machine values for each turn.

        The following values are calculated in this function. All are
        ndarrays of the data type float.

        * time_at_turn
            Time at each turn. Turn zero = 0 [s].
        * phi0
            Synchronous phase angle at the end of each turn.
        * e0
            Total energy of synchronous particle at the end of each turn.
        * beta0
            Lorentz beta factor (v/c) at the end of each turn.
        * deltaE0
            Difference between e0(n) and e0(n-1) for each turn.
        * eta0
            Phase slip factor at each turn.
        * drift_coef
            Coefficient used for calculating difference,
            from phase n to phase n + 1.
            Needed in trajectory height calculator and tracking.
        * omega_rev0
            Revolution frequency at each turn.
        * vrf1_at_turn
            Peak voltage at each turn for the first RF station.
        * vrf2_at_turn
            Peak voltage at each turn for the second RF station.

        The values are saved as fields of the Machine object.
        """
        pass



"""
Module containing functions for pre-processing raw data

The functions here are to be moved over from tomo.data.data_treatment

:Author(s): **Anton Lu**, **Christoffer Hjertø Grindheim**
"""
import typing as t

import numpy as np
from multipledispatch import dispatch
from scipy import optimize

from tomo.utils import physics
from .profiles import Profiles
from .. import assertions as asrt
from ..tracking import Machine
from ..tracking.machine_base import MachineABC


def rebin(waterfall: np.ndarray, rbn: int, dtbin: float = None,
          synch_part_x: float = None) \
        -> t.Union[t.Tuple[np.ndarray, float, float],
                   t.Tuple[np.ndarray, float]]:
    """Rebin waterfall from shape (P, X) to (P, Y).
    P is the number of profiles, X is the original number of bins,
    and Y is the number of bins after the re-binning.

    The algorithm is based on the rebinning function from the
    original tomography program.

    An array of length N, rebinned with a rebin factor of R will result
    in a array of length N / R. If N is not dividable on R, the resulting
    array will have the length (N / R) + 1.

    Parameters
    ----------
    waterfall: ndarray
        2D array of raw-data shaped as waterfall. Shape: (nprofiles, nbins).
    rbn: int
        Rebinning factor.
    dtbin: float
        Size of profile bins [s]. If provided, the function
        will return the new size of the bins after rebinning.
    synch_part_x: float
        x-coordinate of synchronous particle, measured in bins.
        If provided, the function will return its updated coordinate.

    Returns
    -------
    rebinned: ndarray
        Rebinned waterfall
    dtbin: float, optional, default=None
        If a dtbin has been provided in the arguments, the
        new size of the profile bins will be returned. Otherwise,
        None will be returned.
    synch_part_x: float, optional, default=None
        If a synch_part_x has been provided in the arguments, the
        new x-coordinate of the synchronous particle in bins will be returned.
        If unfilled only rebinned and dtbin will be returned
    """
    data = np.copy(waterfall)

    # Check that there is enough data to for the given rebin factor.
    if data.shape[1] % rbn == 0:
        rebinned = _rebin_dividable(data, rbn)
    else:
        rebinned = _rebin_individable(data, rbn)

    if dtbin is not None:
        dtbin *= rbn
    if synch_part_x is not None:
        synch_part_x /= rbn
        return rebinned, dtbin, synch_part_x

    return rebinned, dtbin


@dispatch(np.ndarray, MachineABC)
def fit_synch_part_x(waterfall: np.ndarray, machine: MachineABC) \
        -> t.Tuple[np.ndarray, float, float]:
    """Linear fit to estimate the phase coordinate of the synchronous
    particle. The found phase is returned as a x-coordinate of the phase space
    coordinate systems in fractions of bins. The estimation is done at
    the beam reference profile, which is set in the
    :class:`tomo.tracking.machine.Machine` object.

    Parameters
    ----------
    waterfall : np.ndarray
        2D array of raw-data shaped as waterfall. Shape: (nprofiles, nbins).
    machine : Machine
        Holds all information needed for the particle tracking and the
        generation of initial the particle distribution.

    Returns
    -------
    fitted_synch_part_x
        X coordinate in the phase space coordinate system of the synchronous
        particle given in bin numbers.
    lower bunch limit
        Estimation of the lower bunch limit in bin numbers.
        Needed for :func:`tomo.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.
    upper bunch limit
        Estimation of the upper bunch limit in bin numbers.
        Needed for :func:`tomo.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.

    """

    ref_idx = machine.beam_ref_frame
    ref_prof = waterfall[ref_idx]
    ref_turn = ref_idx * machine.dturns

    # Find the upper and lower tangent for of the bunch.
    # Needed in order to calculate the duration [rad] of the bunch.
    tfoot_up, tfoot_low = _calc_tangentfeet(ref_prof)

    # Calculate the duration of the bunch [rad].
    bunch_duration = (tfoot_up - tfoot_low) * machine.dtbin
    bunch_phaselength = (machine.h_num * bunch_duration
                         * machine.omega_rev0[ref_turn])

    # Estimate the synchronous phase.
    x0 = machine.phi0[ref_turn] - bunch_phaselength / 2.0
    phil = optimize.newton(
        func=physics.phase_low_mch, x0=x0,
        fprime=physics.dphase_low_mch,
        tol=0.0001, maxiter=100,
        args=(machine, bunch_phaselength, ref_turn))

    # Calculates the x coordinate of the synchronous particle given in
    # the phase space coordinate system.
    fitted_synch_part_x = (tfoot_low + (machine.phi0[ref_turn] - phil)
                           / (machine.h_num
                              * machine.omega_rev0[ref_turn]
                              * machine.dtbin))

    return fitted_synch_part_x, tfoot_low, tfoot_up


@dispatch(Profiles)
def fit_synch_part_x(profiles: 'Profiles') -> t.Tuple[np.ndarray, float, float]:
    """Linear fit to estimate the phase coordinate of the synchronous
    particle. The found phase is returned as a x-coordinate of the phase space
    coordinate systems in fractions of bins. The estimation is done at
    the beam reference profile, which is set in the
    :class:`tomo.tracking.machine.Machine` object.

    Parameters
    ----------
    profiles: Profiles
        Profiles object containing waterfall and information about the
        measurement.

    Returns
    -------
    fitted_synch_part_x
        X coordinate in the phase space coordinate system of the synchronous
        particle given in bin numbers.
    lower bunch limit
        Estimation of the lower bunch limit in bin numbers.
        Needed for :func:`tomo.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.
    upper bunch limit
        Estimation of the upper bunch limit in bin numbers.
        Needed for :func:`tomo.utils.tomo_output.write_plotinfo_ftn`
        function in order to write the original output format.

    """
    return fit_synch_part_x(profiles.waterfall, profiles.machine)


def filter_profiles(waterfall: np.ndarray, xp: np.ndarray = None,
                    yp: np.ndarray = None, rec_prof: int = None) \
        -> t.Union[
            np.ndarray,
            t.Tuple[np.ndarray, np.ndarray],
            t.Tuple[np.ndarray, np.ndarray, np.ndarray],
            t.Tuple[np.ndarray, np.ndarray, int],
            t.Tuple[np.ndarray, int]]:
    """
    Filters out empty profiles from measured data. Empty profiles are
    considered just noise. Returned arrays are truncated with the noise frames
    removed.

    Parameters
    ----------
    waterfall: np.ndarray
        Waterfall array of shape (n_profiles, n_bins)
    xp: np.ndarray
        Tracked and binned particles of shape (n_profiles, n_bins)
    yp: np.ndarray
        Tracked and binned particles of shape (n_profiles, n_bins)
    rec_prof: int
        Reconstruction profile. The profile will be shifted by the number of
        removed profiles.

    Returns
    -------
        Variable number of truncated arrays with empty profiles removed,
        depending on the number of inputs.
    """
    if (xp is None and yp is not None) or (xp is not None and yp is None):
        raise ValueError('Both xp and yp must be passed.')

    waterfall = waterfall / waterfall.sum()

    bin_sum = waterfall.sum(axis=1)

    threshold = bin_sum.max() - bin_sum.min() - bin_sum.mean()
    good_frames = bin_sum > threshold

    good_waterfall = waterfall[good_frames, :]

    output = [good_waterfall]
    if xp is not None:
        good_xp = xp[good_frames, :]

        output.append(good_xp)
    if yp is not None:
        good_yp = yp[good_frames, :]

        output.append(good_yp)
    if rec_prof is not None:
        shift = len(waterfall) - len(good_waterfall)
        shifted_rec_prof = rec_prof - shift

        if shifted_rec_prof < 0:
            raise ValueError('Shift of rec_prof will make it negative. '
                             f'rec_prof ({rec_prof}) should be greater or '
                             f'equal to the shift ({shifted_rec_prof}).')

        output.append(shifted_rec_prof)

    return tuple(output)


def cut_waterfall(waterfall: t.Union[t.List[np.ndarray], np.ndarray],
                  cut_left: int, cut_right: int) -> np.ndarray:
    """
    Cut a waterfall array of shape (n_profiles, n_bins) to shape
    (n_profiles, (cut_right - cut_left)) by shaving off the bins
    to the left of cut_left, and to the right of cut_right

    Parameters
    ----------
    waterfall : List[np.ndarray] or np.npdarray
        2D array of raw-data shaped as waterfall. Shape: (nprofiles, nbins).
    cut_left : int
        The left hand side of the cut, counted in bin number
    cut_right :
        The right hands side of the cut, counted in bin number
        Can be negative, and the bin number count is then used from
        the right side of the array.

    Returns
    -------
    np.ndarray :
        A cut waterfall of shape (n_profiles, (cut_right - cut_left))
    """
    if isinstance(waterfall, list):
        waterfall = np.ndarray(waterfall)

    n_bins = len(waterfall[0])
    asrt.assert_greater_or_equal(cut_left, 'cut_left', 0, ValueError)
    if cut_right < 0:
        asrt.assert_inrange(cut_right, 'cut_right', -n_bins, -1, ValueError)
        asrt.assert_greater(n_bins + cut_right, 'cut_right', cut_left,
                            ValueError,
                            extra_text='-cut_right cannot wrap around'
                                       'cut_left')
    else:
        asrt.assert_inrange(cut_right, 'cut_right', 0, n_bins, ValueError,
                            'cut_right cannot be greater than the number'
                            'of bins in the waterfall')
        asrt.assert_greater(cut_right, 'cut_right', cut_left, ValueError,
                            'cut_right must be greater than cut_left')

    waterfall = waterfall[:, cut_left:cut_right]
    return waterfall


# Finds foot tangents of profile. Needed to estimate bunch duration
# when performing a fit to find synch_part_x.
def _calc_tangentfeet(ref_prof: np.ndarray) -> t.Tuple[float, float]:
    nbins = len(ref_prof)
    index_array = np.arange(nbins) + 0.5

    tanbin_up, tanbin_low = _calc_tangentbins(ref_prof, nbins)

    [bl, al] = np.polyfit(index_array[tanbin_low - 2: tanbin_low + 2],
                          ref_prof[tanbin_low - 2: tanbin_low + 2], deg=1)

    [bu, au] = np.polyfit(index_array[tanbin_up - 1: tanbin_up + 3],
                          ref_prof[tanbin_up - 1: tanbin_up + 3], deg=1)

    tanfoot_low = -1 * al / bl
    tanfoot_up = -1 * au / bu

    return tanfoot_up, tanfoot_low


# Returns index of last bins to the left and right of max valued bin,
# with value over the threshold.
def _calc_tangentbins(ref_profile: np.ndarray, nbins: int,
                      threshold_coeff: float = 0.15) -> t.Tuple[float, float]:
    threshold = threshold_coeff * np.max(ref_profile)
    maxbin = np.argmax(ref_profile)
    for ibin in range(maxbin, 0, -1):
        if ref_profile[ibin] < threshold:
            tangent_bin_low = ibin + 1
            break
    for ibin in range(maxbin, nbins):
        if ref_profile[ibin] < threshold:
            tangent_bin_up = ibin - 1
            break

    return tangent_bin_up, tangent_bin_low


# Rebins an 2d array given a rebin factor (rbn).
# The given array MUST have a length equal to an even number.
def _rebin_dividable(data: np.ndarray, rbn: int) -> np.ndarray:
    if data.shape[1] % rbn != 0:
        raise AssertionError('Input array must be '
                             'dividable on the rebin factor.')
    ans = np.copy(data)

    nprofs = data.shape[0]
    nbins = data.shape[1]

    new_nbins = int(nbins / rbn)
    all_bins = new_nbins * nprofs

    ans = ans.reshape((all_bins, rbn))
    ans = np.sum(ans, axis=1)
    ans = ans.reshape((nprofs, new_nbins))

    return ans


# Rebins an 2d array given a rebin factor (rbn).
# The given array MUST have vector length equal to an odd number.
def _rebin_individable(data: np.ndarray, rbn: int) -> np.ndarray:
    nprofs = data.shape[0]
    nbins = data.shape[1]

    ans = np.zeros((nprofs, int(nbins / rbn) + 1))

    last_data_idx = int(nbins / rbn) * rbn
    ans[:, :-1] = _rebin_dividable(data[:, :last_data_idx], rbn)
    ans[:, -1] = _rebin_last(data, rbn)[:, 0]
    return ans


# Rebins last indices of an 2d array given a rebin factor (rbn).
# Needed for the rebinning of odd arrays.
def _rebin_last(data: np.ndarray, rbn: int) -> np.ndarray:
    nprofs = data.shape[0]
    nbins = data.shape[1]

    i0 = (int(nbins / rbn) - 1) * rbn
    ans = np.copy(data[:, i0:])
    ans = np.sum(ans, axis=1)
    ans[:] *= rbn / (nbins - i0)
    ans = ans.reshape((nprofs, 1))
    return ans



"""Module containing the PhaseSpaceInfo class

:Author(s): **Christoffer Hjertø Grindheim**
"""
from typing import TYPE_CHECKING, Tuple

import numpy as np

from tomo.exceptions import EnergyBinningError, EnergyLimitsError, \
    PhaseLimitsError, ArrayLengthError
from .. import assertions as asrt
from ..utils import physics

if TYPE_CHECKING:
    from .machine import Machine


class PhaseSpaceInfo:
    """Class for calculating and storing information about the phase
    space reconstruction area.

    Parameters
    ----------
    machine: Machine
        Object containing machine parameters and settings.

    Attributes
    ----------
    machine: Machine
        Object containing machine parameters and settings.
    jmin: ndarray
        1D array of integers holding the minimum energy for each phase
        of the phase space coordinate system.
    jmax: ndarray
        1D array of integers holding the maximum energy for each phase
        of the phase space coordinate system.
    imin: int
        Minimum phase of reconstruction area, given in phase space coordinates.
    imax: int
        Maximum phase of reconstruction area, given in phase space coordinates.
    dEbin: float
        Energy size of bins in phase space coordinate system.
    xorigin: float
        The absolute difference (in bins) between phase=0 and the origin
        of the reconstructed phase space coordinate system.
    """

    def __init__(self, machine: 'Machine'):
        self.machine = machine
        self.jmin = None
        self.jmax = None
        self.imin = None
        self.imax = None
        self.dEbin = None
        self.xorigin = None

    def find_binned_phase_energy_limits(self):
        """This function finds the limits in the i (phase) and j (energy) axes
        of the reconstructed phase space coordinate system.

        The area within the limits of i and j will be the phase space
        reconstruction area. This is where the particles will be populated.
        By setting the `full_pp_flag` in the provided
        :class:`~tomo.tracking.machine.Machine` object to True,
        the reconstruction area will be set to all of the phase space image.

        This function gives a value to all attributes of the class.

        Raises
        ------
        EnergyBinningError: Exception
            Size of each phase space bin in the energy axis is less than zero.
        PhaseLimitsError: Exception
            Error in the limits of the i (phase) axis.
        EnergyLimitsError: Exception
            Error in the limits of the j (energy) axis.
        ArrayLengthError: Exception
            Difference in length of jmin and jmax.
        """

        self.xorigin = self.calc_xorigin()
        self.dEbin = self.find_dEbin()

        # If dEbin is less than 0, an error is raised.
        asrt.assert_greater(self.dEbin, 'dEbin', 0.0, EnergyBinningError)

        # Is this still a valid choice with the new method?
        if self.machine.full_pp_flag is True:
            (jmin, jmax,
             imin, imax) = self._limits_track_full_image()
        else:
            (jmin, jmax,
             imin, imax) = self._limits_track_rec_area(self.dEbin)

        # Calculate limits (index of bins) in i-axis (phase axis),
        # 	adjust j-axis (energy axis)
        (jmin, jmax,
         self.imin,
         self.imax) = self._adjust_limits(jmax, jmin, imin, imax)

        self.jmin = jmin.astype(np.int32)
        self.jmax = jmax.astype(np.int32)

        # Ensuring that the output is valid
        self._assert_correct_ijlimits()

    def find_dEbin(self) -> np.float64:
        """Function to calculate the size of a energy bin in the
        reconstructed phase space coordinate system.

        Needed by
        :meth:`tomo.tracking.phase_space_info.PhaseSpaceInfo.find_binned_phase_energy_limits`
        in order to calculate the reconstruction area.

        Returns
        -------
        dEbin:
            Energy size of bins in phase space coordinate system.

        Raises
        ------
        EnergyBinningError: Exception
            The provided value of machine.demax is invalid, or xorigin is
            not calculated before the function is called.
        """
        if self.xorigin is None:
            raise EnergyBinningError(
                'xorigin must be calculated in order to find dEbin.')

        turn = self.machine.beam_ref_frame * self.machine.dturns
        phases = self._calculate_phases(turn)
        delta_e_known = 0.0
        asrt.assert_not_equal(self.machine.demax, 'dEmax',
                              0.0, EnergyBinningError,
                              'The specified maximum energy of '
                              'reconstructed phase space is invalid.')
        if self.machine.demax < 0.0:
            if self.machine.vrf2_at_turn[turn] != 0.0:
                energies_low = self._trajectoryheight(
                    phases, phases[0], delta_e_known, turn)

                energies_up = self._trajectoryheight(
                    phases, phases[self.machine.nbins],
                    delta_e_known, turn)

                return (min(np.amax(energies_low), np.amax(energies_up))
                        / (self.machine.nbins - self.machine.synch_part_y))
            else:
                return (self.machine.beta0[turn]
                        * np.sqrt(self.machine.e0[turn]
                                  * self.machine.q
                                  * self.machine.vrf1_at_turn[turn]
                                  * np.cos(self.machine.phi0[turn])
                                  / (2 * np.pi * self.machine.h_num
                                     * self.machine.eta0[turn]))
                        * self.machine.dtbin
                        * self.machine.h_num
                        * self.machine.omega_rev0[turn])
        else:
            return (float(self.machine.demax)
                    / (self.machine.nbins - self.machine.synch_part_y))

    def calc_xorigin(self) -> np.float64:
        """Function for calculating xorigin.

        Needed by
        :func:`~tomo.tracking.phase_space_info.PhaseSpaceInfo.find_binned_phase_energy_limits`
        in order to calculate the reconstruction area.

        Returns
        -------
        xorigin: float
            The absolute difference (in bins) between phase=0 and the origin
            of the reconstructed phase space coordinate system.
        """
        beam_ref_turn = self.machine.beam_ref_frame * self.machine.dturns
        return (self.machine.phi0[beam_ref_turn]
                / (self.machine.h_num
                   * self.machine.omega_rev0[beam_ref_turn]
                   * self.machine.dtbin)
                - self.machine.synch_part_x)

    # Finding limits for distributing particle over the full image
    # of the reconstructed phase space.
    def _limits_track_full_image(self) \
            -> Tuple[np.ndarray, np.ndarray, np.int32, np.int32]:
        jmax = np.zeros(self.machine.nbins)
        jmin = np.copy(jmax)

        jmax[:] = self.machine.nbins
        jmin[:] = np.ceil(2.0 * self.machine.synch_part_y - jmax + 0.5)

        imin = np.int32(0)
        imax = np.int32(self.machine.nbins)
        return jmin, jmax, imin, imax

    # Finding limits for creating a smaller reconstruction area.
    def _limits_track_rec_area(self, dEbin) \
            -> Tuple[np.ndarray, np.ndarray, int, int]:
        jmax = np.zeros(self.machine.nbins)
        jmin = np.copy(jmax)

        # Calculating turn and phases at the filmstart reference point.
        turn = self.machine.filmstart * self.machine.dturns
        phases = self._calculate_phases(turn)

        # Calculate the limits in energy at each phase
        jmax = self._find_max_binned_energy(phases, turn, dEbin)
        jmin = self._find_min_binned_energy(jmax)
        self._assert_jlimits_ok(jmin, jmax)

        # Find area in phase where energy limits are valid
        # (jmax - jmin >= 0).
        imin = self._find_min_binned_phase(jmin, jmax)
        imax = self._find_max_binned_phase(jmin, jmax)

        return jmin, jmax, imin, imax

    # Function for finding maximum energy (j max) for each bin in the profile.
    # The assumption mad in the program is that jmin and jmax are mirrored.
    def _find_max_binned_energy(self, phases: np.ndarray, turn: int,
                                dEbin: np.float64):

        energy = 0.0
        jmax_low = np.zeros(self.machine.nbins + 1)
        jmax_up = np.zeros(self.machine.nbins + 1)

        # finding max energy at edges of profiles
        for i in range(self.machine.nbins + 1):
            temp_energy = np.floor(self.machine.synch_part_y
                                   + self._trajectoryheight(
                                    phases[i], phases[0], energy, turn)
                                   / dEbin)

            jmax_low[i] = int(temp_energy)

            temp_energy = np.floor(self.machine.synch_part_y
                                   + self._trajectoryheight(
                                    phases[i],
                                    phases[self.machine.nbins],
                                    energy, turn)
                                   / dEbin)

            jmax_up[i] = int(temp_energy)

        jmax = np.zeros(self.machine.nbins)
        for i in range(self.machine.nbins):
            jmax[i] = min([jmax_up[i], jmax_up[i + 1],
                           jmax_low[i], jmax_low[i + 1],
                           self.machine.nbins])
        return jmax

    # Function for finding minimum energy (j min) for each bin in profile
    # Checking each element if less than threshold,
    # in such cases will threshold be used.
    def _find_min_binned_energy(self, jmax: np.ndarray, threshold: int = 1) \
            -> np.ndarray:
        jmin = np.ceil(2.0 * self.machine.synch_part_y - jmax[:] - 0.5)
        return np.where(jmin[:] >= threshold, jmin[:], threshold)

    # Finding index for minimum phase for profile
    def _find_min_binned_phase(
            self, jmin: np.ndarray, jmax: np.ndarray) -> int:
        for i in range(0, self.machine.nbins):
            if jmax[i] - jmin[i] >= 0:
                return i

    # Finding index for maximum phase for profile
    def _find_max_binned_phase(
            self, jmin: np.ndarray, jmax: np.ndarray) -> int:
        for i in range(self.machine.nbins - 1, 0, -1):
            if jmax[i] - jmin[i] >= 0:
                return i

    # Adjustment of limits in relation to
    # 	specified input min/max index and found min/max in profile.
    # 	E.g. if profile_mini is greater than allbin_min, use profile_mini.
    # Calculates final limits of i-axis.
    def _adjust_limits(self, jmax: np.ndarray, jmin: np.ndarray,
                       imin: int, imax: int) \
            -> Tuple[np.ndarray, np.ndarray, int, int]:

        # Maximum and minimum bin, as specified by user.
        max_dtbin = int(np.ceil(self.machine.max_dt / self.machine.dtbin))
        min_dtbin = int(self.machine.min_dt / self.machine.dtbin)

        if min_dtbin > imin or self.machine.full_pp_flag:
            imin = min_dtbin
            jmax[:min_dtbin] = np.floor(self.machine.synch_part_y)
            jmin = np.ceil(2.0 * self.machine.synch_part_y - jmax + 0.5)

        if max_dtbin < imax or self.machine.full_pp_flag:
            imax = max_dtbin - 1  # -1 in order to count from idx 0
            jmax[max_dtbin:
                 self.machine.nbins] = np.floor(self.machine.synch_part_y)
            jmin = np.ceil(2.0 * self.machine.synch_part_y - jmax + 0.5)

        return jmin, jmax, imin, imax

    # Returns an array of phases for a given turn
    def _calculate_phases(self, turn: int) -> np.ndarray:
        indarr = np.arange(self.machine.nbins + 1)
        phases = ((self.xorigin + indarr)
                  * self.machine.dtbin
                  * self.machine.h_num
                  * self.machine.omega_rev0[turn])
        return phases

    # Trajectory height calculator
    def _trajectoryheight(self, phi: np.ndarray, phi_known: float,
                          delta_e_known: float, turn: int) -> float:
        machine = self.machine
        if isinstance(machine.phi12, np.ndarray):
            phi12 = machine.phi12[turn]
        else:
            phi12 = machine.phi12

        cplx_height = 2.0 * machine.q / machine.drift_coef[turn]
        cplx_height *= (machine.vrf1_at_turn[turn]
                        * (np.cos(phi) - np.cos(phi_known))
                        + machine.vrf2_at_turn[turn]
                        * (np.cos(machine.h_ratio
                                  * (phi - phi12))
                           - np.cos(machine.h_ratio
                                    * (phi_known - phi12)))
                        / machine.h_ratio
                        + (phi - phi_known)
                        * physics.rf_voltage_at_phase(
                    machine.phi0[turn],
                    machine.vrf1_at_turn[turn],
                    machine.vrf2_at_turn[turn],
                    machine.h_ratio, phi12))
        cplx_height += delta_e_known ** 2

        if np.size(cplx_height) > 1:
            # Returning array
            cplx_height = np.array(cplx_height, dtype=complex)
            cplx_height = np.sqrt(cplx_height)
        else:
            # Returning scalar
            cplx_height = np.sqrt(complex(cplx_height))

        return cplx_height.real

    def _assert_correct_ijlimits(self):
        # Testing imin and imax
        asrt.assert_inrange(self.imin, 'imin', 0, self.imax,
                            PhaseLimitsError,
                            f'imin and imax out of bounds')
        asrt.assert_less_or_equal(self.imax, 'imax', self.jmax.size,
                                  PhaseLimitsError,
                                  f'imin and imax out of bounds')
        # Testing jmin and jmax
        asrt.assert_array_in_range(self.jmin[self.imin:self.imax], 0,
                                   self.jmax[self.imin:self.imax],
                                   EnergyLimitsError,
                                   msg=f'jmin and jmax out of bounds ',
                                   index_offset=self.imin)
        asrt.assert_array_less_eq(self.jmax[self.imin:self.imax],
                                  self.machine.nbins,
                                  EnergyLimitsError,
                                  f'jmin and jmax out of bounds ')
        asrt.assert_equal(self.jmin.shape, 'jmin',
                          self.jmax.shape, ArrayLengthError,
                          'jmin and jmax should have the same shape')
        self._assert_jlimits_ok(self.jmin, self.jmax)

    # Testing that there is a difference between jmin and jmax
    def _assert_jlimits_ok(self, jmin: np.ndarray, jmax: np.ndarray):
        if all(jmin >= jmax):
            raise EnergyLimitsError(
                'All of jmin is larger than or equal to jmax; '
                'the size of the reconstruction area is zero.')



"""Module containing assertion functions with standardised output for
tomography

:Author(s): **Christoffer Hjertø Grindheim**
"""
from numbers import Number
from typing import Union, Type, Tuple, Any, Collection, TYPE_CHECKING

import numpy as np

from tomo import exceptions as expt

if TYPE_CHECKING:
    from tomo.tracking.machine import Machine
    from tomo.utils.tomo_input import Frames


# =========================================================
#                      SCALAR ASSERTIONS
# =========================================================
def assert_greater(var: Union[int, float], var_name: str,
                   limit: Union[int, float], error_class: Type[Exception],
                   extra_text: str = ''):
    """Assert scalar greater than X.

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    limit: int, float
        Limit of which variable should be greater than.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var <= limit:
        msg = _write_std_err_msg(var_name, var, limit, '>', extra_text)
        raise error_class(msg)


def assert_less(
        var: Union[int, float],
        var_name: str, limit: Union[int, float],
        error_class: Type[Exception],
        extra_text: str = ''):
    """Assert scalar less than X.

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    limit: int, float
        Limit of which variable should be less than.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var >= limit:
        msg = _write_std_err_msg(var_name, var, limit, '<', extra_text)
        raise error_class(msg)


def assert_equal(var: Union[int, float], var_name: str,
                 limit: Union[int, float], error_class: Type[Exception],
                 extra_text: str = ''):
    """Assert scalar equal to X.

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    limit: int, float
        Limit of which variable should be equal to.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var != limit:
        msg = _write_std_err_msg(var_name, var, limit, '==', extra_text)
        raise error_class(msg)


def assert_not_equal(var: Union[int, float], var_name: str,
                     limit: Union[int, float], error_class: Type[Exception],
                     extra_text: str = ''):
    """Assert scalar unequal to X.

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    limit: int, float
        Limit of which variable should be unequal to.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var == limit:
        msg = _write_std_err_msg(var_name, var, limit, '!=', extra_text)
        raise error_class(msg)


def assert_less_or_equal(var: Union[int, float], var_name: str,
                         limit: Union[int, float],
                         error_class: Type[Exception], extra_text: str = ''):
    """Assert scalar less than or equal to X.

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    limit: int, float
        Limit of which variable should be less than or equal to.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var > limit:
        msg = _write_std_err_msg(var_name, var, limit, '<=', extra_text)
        raise error_class(msg)


def assert_greater_or_equal(
        var: Union[int, float],
        var_name: str, limit: Union[int, float],
        error_class: Type[Exception],
        extra_text: str = ''):
    """Assert scalar greater than or equal to X.

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    limit: int, float
        Limit of which variable should be greater than or equal to.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var < limit:
        msg = _write_std_err_msg(var_name, var, limit, '>=', extra_text)
        raise error_class(msg)


def assert_inrange(var: Union[int, float], var_name: str,
                   low_lim: Union[int, float], up_lim: Union[int, float],
                   error_class: Type[Exception], extra_text: str = ''):
    """Assert scalar is in range of [x, y].

    Parameters
    ----------
    var: int, float
        Variable to be asserted.
    var_name: string
        Name of variable, for users reference.
    low_lim: int, float
        Lower limit of variable.
    up_lim: int, float
        Upper limit of variable.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if var < low_lim or var > up_lim:
        error_message = (f'\nInput parameter "{var_name}" has the '
                         f'unexpected value: {var}.\n'
                         f'Expected value: {var_name} should be in area '
                         f'[{low_lim}, {up_lim}].')
        error_message += f'\n{extra_text}'
        raise error_class(error_message)


# =========================================================
#                      ARRAY ASSERTIONS
# =========================================================


def assert_array_not_equal(array: np.ndarray, array_name: str,
                           limit: Union[int, float],
                           error_class: Type[Exception], extra_text: str = ''):
    """Assert array not equal to X.

    This function asserts that not all
    elements of a value has a scalar value X.

    Parameters
    ----------
    array: ndarray
        Array to be asserted.
    array_name: string
        Name of array, for user reference.
    limit: int, float
        Scalar value which the array should be unequal of.
    error_class: Exception
        Error class to be raised if assertion fails.
    extra_text: string
        Extra text to be written after standard error message.
    """
    if np.all(array == limit):
        error_message = (f'\nAll elements of the array "{array_name}" '
                         f'has the unexpected value: {array[0]}.\n'
                         f'Expected value: {array_name} != {limit}.')
        error_message += f'\n{extra_text}'
        raise error_class(error_message)


def assert_array_shape_equal(arrays: Tuple[np.ndarray, np.ndarray],
                             array_names: Tuple[str, str],
                             demanded_shape: Tuple[int, int],
                             extra_text: str = ''):
    """Assert that two arrays have a given shape.

    Parameters
    ----------
    arrays: tuple
        tuple should contain (ndarrayX, ndarrayY).
        ndarrays are the arrays to be asserted.
    array_names: tuple
        tuple of strings, being names of the arrays to be tested for the
        users reference.
    demanded_shape: tuple
        Demanded shape of arrays to be asserted. Tuple: (int, int).
    extra_text: string
        Extra text to be written after standard error message.
    """
    assert_greater_or_equal(len(arrays), 'number of arrays', 2,
                            AssertionError,
                            'Unable to compare arrays, since less than two '
                            'arrays are given.')
    ok = True
    counter = 0
    array_error_str = ''
    for i in range(len(arrays)):
        if demanded_shape != arrays[i].shape:
            ok = False
            array_error_str += (f'{array_names[i]} with '
                                f'shape: {arrays[i].shape}\n')
            counter += 1
    if not ok:
        error_message = (f'\ndeviation from a desired array shape '
                         f'of {demanded_shape} found in array(s):\n'
                         f'{array_error_str}'
                         f'{extra_text}')
        raise expt.UnequalArrayShapes(error_message)


def assert_array_numel_equal(arrays: Tuple[np.ndarray, np.ndarray],
                             array_names: Tuple[str, str],
                             demanded_numel: Tuple[int, int],
                             extra_text: str = ''):
    """Assert that two arrays have the given number of items

    Parameters
    ----------
    arrays: tuple
        tuple should contain (ndarrayX, ndarrayY).
        ndarrays are the arrays to be asserted.
    array_names: tuple
        tuple of strings, being names of the arrays to be tested for the
        users reference.
    demanded_numel: tuple
        Demanded shape of arrays to be asserted. Tuple: (int, int).
    extra_text: string
        Extra text to be written after standard error message.
    """
    assert_greater_or_equal(len(arrays), 'number of arrays', 2,
                            AssertionError,
                            'Unable to compare arrays, since less than two '
                            'arrays are given.')
    if isinstance(demanded_numel, Number):
        demanded_numel = (demanded_numel,)

    ok = True
    counter = 0
    array_error_str = ''
    for i in range(len(arrays)):
        if demanded_numel != arrays[i].shape:
            ok = False
            array_error_str += (f'{array_names[i]} with '
                                f'shape: {arrays[i].shape}\n')
            counter += 1
    if not ok:
        error_message = (f'\ndeviation from a desired array size '
                         f'of {demanded_numel} found in array(s):\n'
                         f'{array_error_str}'
                         f'{extra_text}')
        raise expt.UnequalArrayShapes(error_message)


def assert_array_in_range(array: np.ndarray, low_lim: Union[int, float],
                          up_lim: Union[int, float],
                          error_class: Type[Exception],
                          msg: str = '', index_offset: int = 0):
    """Assert all array elements are within a range [x, y].

    Parameters
    ----------
    array: ndarray
        Array to be asserted.
    low_lim: int, float
        Lower limit of array elements.
    up_lim: int, float
        Upper limit of array elements.
    error_class: Exception
        Error class to be raised if assertion fails.
    msg: string
        Extra text to be written after standard error message.
    index_offset: int
        Index offset when checking equality. Only elements after the offset
        are checked.
    """
    log_arr = np.where(np.logical_or(array < low_lim, array > up_lim),
                       False, True)
    _assert_log_arr(log_arr, error_class, index_offset, msg)


def assert_array_greater(array: np.ndarray, limit: Union[int, float],
                         error_class: Type[Exception],
                         msg: str = '', index_offset: int = 0):
    """Assert all array elements are greater than x.

    Parameters
    ----------
    array: ndarray
        Array to be asserted.
    limit: int, float
        Scalar value array elements should be greater than.
    error_class: Exception
        Error class to be raised if assertion fails.
    msg: string
        Extra text to be written after standard error message.
    index_offset: int
        Index offset when checking equality. Only elements after the offset
        are checked.
    """
    log_arr = np.where(array <= limit, False, True)
    _assert_log_arr(log_arr, error_class, index_offset, msg)


def assert_array_greater_eq(array: np.ndarray, limit: Union[int, float],
                            error_class: Type[Exception],
                            msg: str = '', index_offset: int = 0):
    """Assert all array elements are greater than or equal to x.

    Parameters
    ----------
    array: ndarray
        Array to be asserted.
    limit: int, float
        Scalar value array elements should be greater than or equal to.
    error_class: Exception
        Error class to be raised if assertion fails.
    msg: string
        Extra text to be written after standard error message.
    index_offset: int
        Index offset when checking equality. Only elements after the offset
        are checked.
    """
    log_arr = np.where(array < limit, False, True)
    _assert_log_arr(log_arr, error_class, index_offset, msg)


def assert_array_less(array: np.ndarray, limit: Union[int, float],
                      error_class: Type[Exception],
                      msg: str = '', index_offset: int = 0):
    """Assert all array elements are less than x.

    Parameters
    ----------
    array: ndarray
        Array to be asserted.
    limit: int, float
        Scalar value array elements should be less than.
    error_class: Exception
        Error class to be raised if assertion fails.
    msg: string
        Extra text to be written after standard error message.
    index_offset: int
        Index offset when checking equality. Only elements after the offset
        are checked.
    """
    log_arr = np.where(array >= limit, False, True)
    _assert_log_arr(log_arr, error_class, index_offset, msg)


def assert_array_less_eq(array: np.ndarray, limit: Union[int, float],
                         error_class: Type[Exception],
                         msg: str = '', index_offset: int = 0):
    """Assert all array elements are less than or equal to x.

    Parameters
    ----------
    array: ndarray
        Array to be asserted.
    limit: int, float
        Scalar value array elements should be less than or equal to.
    error_class: Exception
        Error class to be raised if assertion fails.
    msg: string
        Extra text to be written after standard error message.
    index_offset: int
        Index offset when checking equality. Only elements after the offset
        are checked.
    """
    log_arr = np.where(array > limit, False, True)
    _assert_log_arr(log_arr, error_class, index_offset, msg)


# Checks that all elements of a logical array is true.
# If not, an error is raised.
# Used by the array assertion functions.
def _assert_log_arr(log_array_ok: np.ndarray, error_class: Type[Exception],
                    index_offset: int, msg: str):
    if not log_array_ok.all():
        error_msg = '\nError found at index: ' \
                    '{}\n'.format(
                        np.argwhere(log_array_ok is False).flatten()
                        + index_offset
                    )
        raise error_class(error_msg + msg)


# =========================================================
#                 TRACKING/PARTICLE ASSERTIONS
# =========================================================

def assert_only_valid_particles(xp: np.ndarray, n_bins: int, msg: str = ''):
    """Assert all particles are within the image width.

    An InvalidParticleError is raised if the trajectory
    of one or more particles goes outside of the image width.

    Parameters
    ----------
    xp: ndarray
        Array of particles.
    n_bins: int
        Number of bins in a profile measurement (image width).
    msg: string
        Extra text for error message.

    Raises
    ------
    InvalidParticleError: Exception
        One or mor particles has left the image.
    """
    if np.any(np.logical_or(xp >= n_bins, xp < 0)):
        err_msg = f'Invalid (lost) particle(s) was found in xp\n'
        raise expt.InvalidParticleError(err_msg + msg)


# =========================================================
#                     MACHINE ASSERTIONS
# =========================================================


def assert_fields(obj: Any, obj_name: str, needed_fields: Collection,
                  error_class: Type[Exception], msg: str = ''):
    """Assert that object contains all necessary fields.

    Parameters
    ----------
    obj: Object
        Object to be asserted.
    obj_name: string
        Name of object, for user reference.
    needed_fields: array like
        Array containing attributes which object should contain.
    error_class: Exception
        Error class to be raised if assertion fails.
    msg: string
        Extra text for error message.
    """
    for field in needed_fields:
        if not hasattr(obj, field) or getattr(obj, field) is None:
            err_msg = f'Missing parameter "{field}" in {obj_name}.'
            if len(msg) > 0:
                err_msg += f'\n{msg}'
            raise error_class(err_msg)


def assert_machine_input(machine: 'Machine'):
    """Assert that input parameters for a machine object is valid.

    Parameters
    ----------
    machine: Machine
        Machine object to be asserted.

    Raises
    ------
    MachineParameterError: Exception
        Invalid value found for machine parameter.
    SpaceChargeParameterError: Exception
        Invalid value found for machine parameter regarding self-field
        measurements.
    """
    # Bin assertions
    assert_var_not_none(machine.dtbin, 'dtbin', expt.MachineParameterError)
    assert_greater(machine.dtbin, 'dtbin', 0, expt.MachineParameterError,
                   'NB: dtbin is the difference of time in bin')

    assert_var_not_none(machine.dturns, 'dturns', expt.MachineParameterError)
    assert_greater(machine.dturns, 'dturns', 0, expt.MachineParameterError,
                   'NB: dturns is the number of machine turns'
                   'between each measurement')

    # Assertions: profile to be reconstructed
    if (machine.filmstart is None or machine.filmstop is None or
            machine.filmstep is None):
        raise expt.MachineParameterError(
            'film start, film stop and filmstep cannot be None, but '
            'must be positive integers, within the number of profiles.')
    else:
        assert_greater_or_equal(machine.filmstart, 'film start',
                                0, expt.MachineParameterError)
        assert_greater_or_equal(machine.filmstop, 'film stop',
                                machine.filmstart, expt.MachineParameterError)
        assert_less_or_equal(abs(machine.filmstep), 'film step',
                             abs(machine.filmstop - machine.filmstart + 1),
                             expt.MachineParameterError)
        assert_not_equal(machine.filmstep, 'film step', 0,
                         expt.MachineParameterError)

    # Reconstruction parameter assertions
    assert_var_not_none(machine.niter, 'niter', expt.MachineParameterError)
    assert_greater(machine.niter, 'niter', 0, expt.MachineParameterError,
                   'NB: niter is the number of iterations of the '
                   'reconstruction process')

    assert_var_not_none(machine.snpt, 'snpt', expt.MachineParameterError)
    assert_greater(machine.snpt, 'snpt', 0, expt.MachineParameterError,
                   'NB: snpt is the square root '
                   'of number of tracked particles in each cell'
                   'of the reconstructed phase space.')

    # Reference frame assertions
    assert_var_not_none(machine.machine_ref_frame, 'Machine ref. frame',
                        expt.MachineParameterError)
    assert_greater_or_equal(machine.machine_ref_frame, 'Machine ref. frame',
                            0, expt.MachineParameterError)
    assert_var_not_none(machine.beam_ref_frame, 'Beam ref. frame',
                        expt.MachineParameterError)
    assert_greater_or_equal(machine.beam_ref_frame, 'Beam ref. frame',
                            0, expt.MachineParameterError)

    # Machine parameter assertion
    assert_var_not_none(
        machine.h_num, 'Harmonic number', expt.MachineParameterError)
    assert_greater_or_equal(
        machine.h_num, 'Harmonic number', 1, expt.MachineParameterError)

    assert_var_not_none(machine.h_ratio, 'Harmonic ratio',
                        expt.MachineParameterError)
    assert_greater_or_equal(machine.h_ratio, 'Harmonic ratio',
                            1, expt.MachineParameterError)

    assert_var_not_none(machine.b0, 'B field (B0)', expt.MachineParameterError)
    assert_greater(machine.b0, 'B field (B0)', 0, expt.MachineParameterError)

    assert_var_not_none(machine.mean_orbit_rad, 'mean orbit radius',
                        expt.MachineParameterError)
    assert_greater(machine.mean_orbit_rad, 'mean orbit radius',
                   0, expt.MachineParameterError)
    assert_var_not_none(machine.bending_rad, 'Bending radius',
                        expt.MachineParameterError)
    assert_greater(machine.bending_rad, 'Bending radius',
                   0, expt.MachineParameterError)
    assert_var_not_none(machine.e_rest, 'rest energy',
                        expt.MachineParameterError)
    assert_greater(machine.e_rest, 'rest energy',
                   0, expt.MachineParameterError)

    # Space charge parameter assertion
    if machine.pickup_sensitivity is not None:
        assert_greater_or_equal(machine.pickup_sensitivity,
                                'pick-up sensitivity',
                                0, expt.SpaceChargeParameterError)
    if machine.g_coupling is not None:
        assert_greater_or_equal(machine.g_coupling, 'g_coupling',
                                0, expt.SpaceChargeParameterError,
                                'NB: g_coupling:'
                                'geometrical coupling coefficient')


def assert_frame_inputs(frame: 'Frames'):
    """Assert that frame parameters are valid, and that raw data will be
    correctly shaped to waterfall.

    Parameters
    ----------
    frame: Frame
        Frame object to be asserted.

    Raises
    ------
    InputError: Exception
        An invalid frame parameter has been found.
    """
    assert_greater(frame.nframes, 'nr of frames', 0, expt.InputError)
    assert_inrange(frame.skip_frames, 'skip frames',
                   0, frame.nframes - 1, expt.InputError)
    assert_greater(frame.nbins_frame, 'frame length', 0, expt.InputError)
    assert_inrange(frame.skip_bins_start, 'skip bins start',
                   0, frame.nbins_frame - 1, expt.InputError,
                   'The number of skipped bins in for a frame cannot'
                   'exceed the number of bins in that same frame')
    assert_inrange(frame.skip_bins_end, 'skip bins end',
                   0, frame.nbins_frame - 1, expt.InputError,
                   'The number of skipped bins in for a frame cannot'
                   'exceed the number of bins in that same frame')
    assert_less(frame.skip_bins_start + frame.skip_bins_end,
                'total bins skipped', frame.nbins_frame, expt.InputError,
                'The number of skipped bins in for a frame cannot'
                'exceed the number of bins in that same frame')
    assert_greater_or_equal(frame.rebin, 're-binning factor',
                            1, expt.InputError)


# =========================================================
#                      INDEX ASSERTIONS
# =========================================================

def assert_index_ok(index: int, index_limit: int, wrap_around: bool = False) \
        -> int:
    """Assert index is valid.

    Assert that index is within bounds [0, index limit).
    Wrap_around for negative indices can be enabled. If this is the
    case, index -1 is points at the last index.

    Parameters
    ----------
    index: int
        Index to be tested.
    index_limit: int
        Maximum index + 1
    wrap_around: bool, optional, default=False
        Set to true to enable wrap around for negative indices.

    Returns
    -------
    index: int
        Asserted index from 0 to index_limit-1.

    Raises
    ------
    IndexError: Exception
        Index out of bounds.
    NegativeIndexError: Exception
        Negative index given without setting wrap_around=True
    """
    index = int(index)
    if index < 0:
        if not wrap_around:
            raise expt.NegativeIndexError('Index cannot be negative')
        elif index_limit + index < 0:
            raise IndexError('Index is out of bounds')
        else:
            index += index_limit
    else:
        if index >= index_limit:
            raise IndexError('Index is out of bounds')
    return index


# =========================================================
#                     ASSERTION UTILITIES
# =========================================================

# Generate standard error message for asserting scalars.
def _write_std_err_msg(var_name: str, var: Union[int, float],
                       limit: Union[int, float],
                       operator: str, extra: str) -> str:
    error_message = (f'\nInput parameter "{var_name}" has the '
                     f'unexpected value: {var}.\n'
                     f'Expected value: {var_name} {operator} {limit}.')
    error_message += f'\n{extra}'
    return error_message


def assert_var_not_none(var: Any, var_name: str, error_class: Type[Exception]):
    if var is None:
        raise error_class(f'{var_name} cannot be of type None')



from argparse import ArgumentParser

from .utils.tomo_run import run


if __name__ == '__main__':
    parser = ArgumentParser()

    parser.add_argument('input', default='stdin', type=str,
                        help='Path to input .dat file, or use "stdin" for '
                             'taking input from stdin.')
    parser.add_argument('-t', '--tomoscope', default=False,
                        action='store_true', help='Run in Tomoscope mode.')
    parser.add_argument('-o', '--output', type=str,
                        help='Output directory for saving stuff.')
    parser.add_argument('-p', '--plot', action='store_true', default=False,
                        help='Plot phase space after reconstruction.')

    args = parser.parse_args()

    run(args.input, tomoscope=args.tomoscope, output_dir=args.output,
        plot=args.plot)



from setuptools import setup
from os import path
import platform
from pybind11.setup_helpers import Pybind11Extension

HERE = path.split(path.abspath(__file__))[0]

extra_compile_args = []
extra_link_args = []
if platform.system() == 'Windows':
    extra_compile_args.append('-openmp')
elif platform.system() == 'Linux':
    extra_compile_args.append('-fopenmp')
    extra_compile_args.append('-ffast-math')
    extra_link_args.append('-lgomp')


cpp_routines = Pybind11Extension('tomo.cpp_routines.libtomo',
                                 cxx_std=17,
                                 sources=['tomo/cpp_routines/libtomo.cpp',
                                          'tomo/cpp_routines/reconstruct.cpp',
                                          'tomo/cpp_routines/kick_and_drift.cpp',
                                          'tomo/cpp_routines/data_treatment.cpp'],
                                 extra_compile_args=extra_compile_args,
                                 extra_link_args=extra_link_args)

setup(ext_modules=[cpp_routines])



"""Module containing the Tracking class.

:Author(s): **Christoffer Hjertø Grindheim**
"""
from typing import Tuple, TYPE_CHECKING, Callable

import numpy as np
import logging

from .. import assertions as asrt
from .__tracking import ParticleTracker
from ..cpp_routines import libtomo
from ..compat import fortran

if TYPE_CHECKING:
    from .machine import Machine

log = logging.getLogger(__name__)


class Tracking(ParticleTracker):
    """Class for particle tracking.

    This class perform the particle tracking based on the algorithm from
    the original tomography program. Here, an initial distribution of test
    particles are homogeneously distributed across the reconstruction area of
    the phase space image. Later, the particles will be tracked trough
    all machine turns and saved for every time frame.

    Parameters
    ----------
    machine: Machine
        Holds all information needed for the particle tracking and the
        generation of initial the particle distribution.

    Attributes
    ----------
    machine: Machine
        Holds all information needed for the particle tracking and the
        generation of initial the particle distribution.
    particles: Particles
        Creates and/or stores the initial particle distribution.
    nturns: int
        Number of machine turns particles should be tracked trough.
    self_field_flag: boolean
        Flag to indicate is self-fields should be included during tracking.
    fortran_flag: boolean
        Flag to indicate is a Fortran-style output should be printed to
        stdout during particle tracking.
    """

    def __init__(self, machine: 'Machine'):
        super().__init__(machine)

    def track(self, recprof: int, init_distr: Tuple[float, float] = None,
              callback: Callable = None) \
            -> Tuple[np.ndarray, np.ndarray]:
        """Primary function for tracking particles.

        The tracking routine starts at a given time frame, with an initial
        distribution of particles. From here, the particles are tracked
        'forward' towards the last time frame and 'backwards' towards the
        first time frame.

        By default, an distribution of particles is spread out homogeneously
        over the area to be reconstructed. This area is found using the
        :class:`tomo.tracking.phase_space_info.PhaseSpaceInfo` class.
        The homogeneous distribution is placed on the time frame intended
        to be reconstructed for optimum quality. This is based on the
        original tomography algorithm.

        An user specified distribution can be given and override the
        default, automatic generation of particles.

        By calling
        :func:`tomo.tracking.__tracking.ParticleTracker.enable_self_fields`,
        a flag indicating that self-fields should be included is set.
        In this case, :func:`kick_and_drift_self` will be used.
        Note that tracking including self fields are much slower than without.

        By calling
        :func:`tomo.tracking.__tracking.ParticleTracker.enable_fortran_output`,
        an output resembling the original is written to stdout.
        Note that the values for the number of lost particles is **not valid**.
        Note also, that if the full Fortran output is to be printed,
        the automatic generation of particles must be performed.

        Parameters
        ----------
        recprof: int
            The profile (time frame) to be reconstructed. Here, the particles
            will have its initial distribution. By giving negative values as
            arguments, the index will count from the last time frame.
        init_distr: tuple, optional, default=None
            An user generated initial distribution. Must be given as a tuple of
            coordinates (dphi, denergy). dphi is the phase difference
            from the synchronous particle [rad]. denergy is the difference
            in energy from the synchronous particle [eV].
        callback: Callable
            Passing a callback with function signature
            (progress: int, total: int) will allow the tracking loop to call
            this function at the end of each turn, allowing the python caller
            to monitor the progress.

        Returns
        -------
        xp: ndarray
            2D array containing each particles x-coordinate at
            each time frame. Shape: (nprofiles, nparts).

            * If self-fields are enabled,
              the coordinates will be given as phase-space coordinates.
            * If self-fields are disabled, the returned x-coordinates will be
              given as phase [rad] relative to the synchronous particle.

        yp: ndarray, float
            2D array containing each particles y-coordinate at
            each time frame. Shape: (nprofiles, nparts).

            * If self-fields are enabled,
              the coordinates will be given as phase-space coordinates.
            * If self-fields are disabled, the returned y-coordinates will be
              given as energy [eV] relative to the synchronous particle.
        """

        recprof = asrt.assert_index_ok(
            recprof, self.machine.nprofiles, wrap_around=True)
        machine = self.machine

        if init_distr is None:
            # Homogeneous distribution is created based on the
            # original Fortran algorithm.
            log.info('Creating homogeneous distribution of particles.')
            self.particles.homogeneous_distribution(machine, recprof)
            coords = self.particles.coordinates_dphi_denergy

            # Print fortran style plot info. Needed for tomograph.
            if self.fortran_flag:
                print(fortran.write_plotinfo(
                    machine, self.particles, self._profile_charge))

        else:
            log.info('Using initial particle coordinates set by user.')
            self.particles.coordinates_dphi_denergy = init_distr
            coords = self.particles.coordinates_dphi_denergy

        dphi = coords[0]
        denergy = coords[1]

        rfv1 = machine.vrf1_at_turn * machine.q
        rfv2 = machine.vrf2_at_turn * machine.q

        # Tracking particles
        if self.self_field_flag:
            # Tracking with self-fields
            log.info('Tracking particles... (Self-fields enabled)')
            denergy = np.ascontiguousarray(denergy)
            dphi = np.ascontiguousarray(dphi)
            xp, yp = self.kick_and_drift_self(
                denergy, dphi, rfv1, rfv2, recprof)
        else:
            # Tracking without self-fields
            nparts = len(dphi)
            nturns = machine.dturns * (machine.nprofiles - 1)
            xp = np.zeros((machine.nprofiles, nparts))
            yp = np.zeros((machine.nprofiles, nparts))

            # Calling C++ implementation of tracking routine.
            libtomo.kick_and_drift(xp, yp, denergy, dphi, rfv1, rfv2,
                                   machine.phi0, machine.deltaE0,
                                   machine.drift_coef, machine.phi12,
                                   machine.h_ratio, machine.dturns,
                                   recprof, nturns, nparts,
                                   self.fortran_flag, callback=callback)

        log.info('Tracking completed!')
        return xp, yp

    def kick_and_drift(self, denergy: np.ndarray, dphi: np.ndarray,
                       rf1v: np.ndarray, rf2v: np.ndarray, rec_prof: int) -> \
            Tuple[np.ndarray, np.ndarray]:
        """Routine for tracking a distribution of particles for N turns.
        N is given by *tracking.nturns*

        A full C++ implementation is used in :func:`track`. This function is
        implemented as hybrid between Python and C++, and kept for reference.

        Parameters
        ----------
        denergy: ndarray
            particle energy relative to synchronous particle [eV]
        dphi: ndarray
            particle phase relative to synchronous particle [rad]
        rf1v: ndarray
            Radio frequency voltages (RF station 1),
            multiplied by particle charge
        rf2v: ndarray
            Radio frequency voltages (RF station 2),
            multiplied by particle charge
        rec_prof: int
            Time slice to initiate tracking.

        Returns
        -------
        dphi:
            Phase [rad] relative to the synchronous particle, for each particle
            at each measured time frame. Shape: (nprofiles, nparts).
        denergy: ndarray
            Energy [eV] relative to the synchronous particle, for each particle
            at each measured time slice [eV]. Shape: (nprofiles, nparts).
        """
        nparts = len(denergy)

        # Creating arrays for all tracked particles
        out_dphi = np.zeros((self.machine.nprofiles, nparts))
        out_denergy = np.copy(out_dphi)

        # Setting homogeneous coordinates to profile to be reconstructed.
        out_dphi[rec_prof] = np.copy(dphi)
        out_denergy[rec_prof] = np.copy(denergy)

        rec_turn = rec_prof * self.machine.dturns
        turn = rec_turn
        profile = rec_prof

        # Tracking 'upwards'
        while turn < self.nturns:
            # Calculating change in phase for each particle at a turn
            dphi = libtomo.drift(denergy, dphi, self.machine.drift_coef,
                                 nparts, turn)
            turn += 1
            # Calculating change in energy for each particle at a turn
            denergy = libtomo.kick(self.machine, denergy, dphi, rf1v, rf2v,
                                   nparts, turn)

            if turn % self.machine.dturns == 0:
                profile += 1
                out_dphi[profile] = np.copy(dphi)
                out_denergy[profile] = np.copy(denergy)
                if self.fortran_flag:
                    fortran.print_tracking_status(rec_prof, profile)

        # Starting again from homogenous distribution
        dphi = np.copy(out_dphi[rec_prof])
        denergy = np.copy(out_denergy[rec_prof])
        turn = rec_turn
        profile = rec_prof

        # Tracking 'downwards'
        while turn > 0:
            # Calculating change in energy for each particle at a turn
            denergy = libtomo.kick(self.machine, denergy, dphi, rf1v, rf2v,
                                   nparts, turn, up=False)
            turn -= 1
            # Calculating change in phase for each particle at a turn
            dphi = libtomo.drift(denergy, dphi, self.machine.drift_coef,
                                 nparts, turn, up=False)

            if turn % self.machine.dturns == 0:
                profile -= 1
                out_dphi[profile] = np.copy(dphi)
                out_denergy[profile] = np.copy(denergy)
                if self.fortran_flag:
                    fortran.print_tracking_status(rec_prof, profile)

        return out_dphi, out_denergy

    def kick_and_drift_self(self, denergy: np.ndarray, dphi: np.ndarray,
                            rf1v: np.ndarray, rf2v: np.ndarray,
                            rec_prof: int) -> Tuple[np.ndarray, np.ndarray]:
        """Routine for tracking a given distribution of particles,\
        including self-fields. Implemented as hybrid between Python and C++.

        Routine for tracking, with self-fields, a distribution of
        particles for N turns. N is given by *tracking.nturns*.

        Used by the function :py:meth:`tomo.tracking.tracking.Tracking.track`
        to track using self-fields.

        Returns the coordinates of the particles as phase space coordinates
        for efficiency reasons due to tracking algorithm based on the
        Fortran tomography. large room for improvement.

        Fortran output not yet supported.

        Parameters
        ----------
        denergy: ndarray
            particle energy relative to synchronous particle [eV]
        dphi: ndarray
            particle phase relative to synchronous particle [rad]
        rf1v: ndarray
            Radio frequency voltages (RF station 1),
            multiplied by particle charge
        rf2v: ndarray
            Radio frequency voltages (RF station 2),
            multiplied by particle charge
        rec_prof: int
            Time slice to initiate tracking.

        Returns
        -------
        xp: ndarray, float
            2D array holding the x-coordinates of each particles at
            each time frame (nprofiles, nparts). Coordinates given
            in bins of phase space coordinate system.
        yp: ndarray, float
            2D array holding the y-coordinates of each particles at
            each time frame (nprofiles, nparts). Coordinates given
            in bins of phase space coordinate system.
        """
        nparts = len(denergy)

        # Creating arrays for all tracked particles
        xp = np.zeros((self.machine.nprofiles, nparts))
        yp = np.copy(xp)

        rec_turn = rec_prof * self.machine.dturns
        turn = rec_turn
        profile = rec_prof

        # To be saved for downwards tracking
        dphi0 = np.copy(dphi)
        denergy0 = np.copy(denergy)

        xp[rec_prof] = self._calc_xp_sf(
            dphi, self.machine.phi0[rec_turn],
            self.particles.xorigin, self.machine.h_num,
            self.machine.omega_rev0[rec_turn],
            self.machine.dtbin, self._phiwrap)

        yp[rec_prof] = (denergy / self.particles.dEbin
                        + self.machine.synch_part_y)

        if self.fortran_flag:
            fortran.print_tracking_status(rec_prof, profile)
        while turn < self.nturns:
            dphi = libtomo.drift(denergy, dphi, self.machine.drift_coef,
                                 nparts, turn)

            turn += 1

            temp_xp = self._calc_xp_sf(dphi, self.machine.phi0[turn],
                                       self.particles.xorigin,
                                       self.machine.h_num,
                                       self.machine.omega_rev0[turn],
                                       self.machine.dtbin,
                                       self._phiwrap)
            selfvolt = self._vself[profile, temp_xp.astype(int) - 1]

            denergy = libtomo.kick(self.machine, denergy, dphi, rf1v, rf2v,
                                   nparts, turn)
            denergy += selfvolt * self.machine.q

            if turn % self.machine.dturns == 0:
                profile += 1
                xp[profile] = temp_xp
                yp[profile] = (denergy / self.particles.dEbin
                               + self.machine.synch_part_y)
                if self.fortran_flag:
                    fortran.print_tracking_status(rec_prof, profile)

        dphi = dphi0
        denergy = denergy0
        turn = rec_turn
        profile = rec_prof - 1
        temp_xp = xp[rec_prof]

        while turn > 0:
            selfvolt = self._vself[profile, temp_xp.astype(int)]

            denergy = libtomo.kick(self.machine, denergy, dphi, rf1v, rf2v,
                                   nparts, turn, up=False)

            denergy += selfvolt * self.machine.q

            turn -= 1

            dphi = libtomo.drift(denergy, dphi, self.machine.drift_coef,
                                 nparts, turn, up=False)

            temp_xp = self._calc_xp_sf(
                dphi, self.machine.phi0[turn], self.particles.xorigin,
                self.machine.h_num, self.machine.omega_rev0[turn],
                self.machine.dtbin, self._phiwrap)

            if turn % self.machine.dturns == 0:
                xp[profile] = temp_xp
                yp[profile] = (denergy / self.particles.dEbin
                               + self.machine.synch_part_y)
                profile -= 1
                if self.fortran_flag:
                    fortran.print_tracking_status(rec_prof, profile)

        return xp, yp

    # Calculate from physical coordinates to x-coordinates.
    # Needed for tracking using self-fields.
    # TODO: removed njit, reimplement in C in the future
    @staticmethod
    def _calc_xp_sf(dphi: np.ndarray, phi0: np.ndarray, xorigin: int, h_num,
                    omega_rev0: np.ndarray, dtbin: int, phiwrap: float) \
            -> np.ndarray:
        temp_xp = (dphi + phi0 - xorigin * h_num * omega_rev0 * dtbin)
        temp_xp = ((temp_xp - phiwrap * np.floor(temp_xp / phiwrap))
                   / (h_num * omega_rev0 * dtbin))
        return temp_xp
